"""
Mediation Analysis

Interfaces with DLATK and Statsmodels

"""
import sys
import numpy as np
import pandas as pd
import statsmodels.api as sm
import scipy.stats as st
import csv
import collections
import patsy
from math import sqrt
from scipy.stats import zscore
from .dlaConstants import pCorrection, DEF_P, DEF_MAX_MED_SUMMARY_SIZE, warn
import itertools

class MediationAnalysis:
	"""
	Interface between Mediation class in Statsmodels and DLATK with the addition of standard Baron and Kenny approach. 

	Parameters
	----------
	outcomeGetter : OutcomeGetter object
	featureGetter : FeatureGetter object
	pathStartNames : list 
	mediatorNames : list 
	outcomeNames : list 
	controlNames : list 
	mediation_method : str 
		"parametric" or "bootstrap" 
	boot_number : int 
		number of bootstrap iterations
	sig_level : float
		significane level for reporting results in summary 
	

	

	Attributes
	----------
	output : dict 
	output_sobel : dict 
	output_p : dict
	baron_and_kenny : boolean
		if True runs Baron and Kenny method
	imai_and_keele : boolean
		if True runs Imai, Keele, and Tingley method


	"""

	def __init__(self, fg, og, path_starts, mediators, outcomes, controls, method="parametric", boot_number=1000, sig_level=DEF_P, style='baron'):
		
		self.outcomeGetter = og
		self.featureGetter = fg

		self.pathStartNames = path_starts  
		self.mediatorNames = mediators
		self.outcomeNames = outcomes  
		self.controlNames = controls

		self.mediation_method = method
		self.boot_number = boot_number
		self.sig_level = sig_level

		self.output = dict()
		self.output_sobel = dict()
		self.output_p = dict() # [c_p, c'_p, alpha_p, beta_p, sobel_p, ...]

		if style == 'baron':
			self.baron_and_kenny = True
			self.imai_and_keele = False
		else:
			try:
				from statsmodels import Mediation
				if style == 'imai':
					self.baron_and_kenny = False
					self.imai_and_keele = True
				elif style == 'both':
					self.baron_and_kenny = True
					self.imai_and_keele = True
			except ImportError:
				warn_string = "Statsmodels Mediation package is currently available only via github. Please manually install this module."
				if style == 'both':
					self.baron_and_kenny = True
					self.imai_and_keele = False
					warn(warn_string + "\nContinuing with Baron and Kenny.", attention=True)
				else:
					warn(warn_string + "\nUnless installed mediation can only be run via Baron and Kenny (in dlatkInterface.py: --mediation_method baron)")
					sys.exit()


	def _truncate_groups(seq, max_group_size, key):
		"""yield only the first `max_group_size` elements from each sub-group of `seq`"""
		for key, group in itertools.groupby(seq, key):
			for item in list(group)[:max_group_size]:
				yield item

	def print_summary(self, output_name=''):
		summary_results = []
		if output_name:
			if ".csv" in output_name:
				csv_name = output_name.replace(".csv", "_summary.csv")
			else:
				csv_name = output_name + "_summary.csv"
		else:
			csv_name = "mediation_summary.csv"
				
		header = ["Path Start", "Outcome", "Mediator"]
		if self.baron_and_kenny:
			header = header + ["c-c'", "sobel_P", "alpha", "beta", "c'"]
		if self.imai_and_keele: 
			header = header + ["Prop_mediated_average_Estimate", "Prop_mediated_average_P_value", 
				"ACME_average_Estimate", "ACME_average_P_value"] 

		for path_start in self.output:
			for outcome in self.output[path_start]:
				count = 0
				for mediator in self.output[path_start][outcome]:
					sobel_p = self.output_p[path_start][outcome][mediator][4]
					if sobel_p <= self.sig_level:
						results = [path_start, outcome, mediator]
						if self.baron_and_kenny:
							results = results + [self.output_sobel[path_start][outcome][mediator].tolist()[4] , sobel_p] + \
								[self.output_sobel[path_start][outcome][mediator].tolist()[6], 
								self.output_sobel[path_start][outcome][mediator].tolist()[9], 
								self.output_sobel[path_start][outcome][mediator].tolist()[2]]
						if self.imai_and_keele: 
							results = results + [self.output[path_start][outcome][mediator][0], 
								self.output_p[path_start][outcome][mediator][5],
								self.output[path_start][outcome][mediator][4],
								self.output_p[path_start][outcome][mediator][6]]
						summary_results.append(results)
							

		summary_results.sort(key=lambda x: (x[0].lower(), x[1].lower(), -abs(x[3])), reverse=False)
		if len(summary_results) > 0:
			print("Printing results to: %s" % csv_name)
			with open(csv_name, 'w', newline='') as csvfile:
				f = csv.writer(csvfile, quotechar='"', quoting=csv.QUOTE_NONNUMERIC, delimiter=',')
				f.writerow(header)
				for key, group in itertools.groupby(summary_results, key=lambda x: (x[0], x[1])):
					for item in list(group)[:DEF_MAX_MED_SUMMARY_SIZE]:
						f.writerow(item)
		else:
			warn("Summary: nothing passes significance threshold of %s." % (self.sig_level))
			

	def print_csv(self, output_name=''):
		if output_name:
			csv_name = output_name
		else:
			csv_name = "mediation.csv"
				
		header = ["Path Start", "Outcome", "Mediator"]
		if self.baron_and_kenny:
			header = header + ["c", "c_p", "c'", "c'_p", "c-c'", "alpha*beta", "alpha", "alpha_error", "alpha_p", 
				"beta", "beta_error", "beta_p", "sobel", "sobel_SE", "sobel_P"]
		if self.imai_and_keele: 
			header = header + ["Prop_mediated_average_Estimate", "Prop_mediated_average_P_value", 
				"ACME_average_Estimate", "ACME_average_P_value", 
				"ADE_average_Estimate", "ADE_average_P_value", 
				"Prop_mediated_average_Lower_CI_bound", "Prop_mediated_average_Upper_CI_bound", 
				"ACME_average_Lower_CI_bound", "ACME_average_Upper_CI_bound", 
				"ADE_average_Lower_CI_bound", "ADE_average_Upper_CI_bound", 
				"ACME_treated_Estimate", "ACME_treated_P_value", "ACME_treated_Lower_CI_bound", "ACME_treated_Upper_CI_bound", 
				"ACME_control_Estimate", "ACME_control_P_value", "ACME_control_Lower_CI_bound", "ACME_control_Upper_CI_bound", 
				"ADE_treated_Estimate", "ADE_treated_P_value","ADE_treated_Lower_CI_bound", "ADE_treated_Upper_CI_bound", 
				"ADE_control_Estimate", "ADE_control_P_value", "ADE_control_Lower_CI_bound", "ADE_control_Upper_CI_bound",
				"Total_effect_Estimate", "Total_effect_P_value", "Total_effect_Lower_CI_bound", "Total_effect_Upper_CI_bound", 
				"Prop_mediated_control_Estimate", "Prop_mediated_control_P_value", "Prop_mediated_control_Lower_CI_bound", "Prop_mediated_control_Upper_CI_bound", 
				"Prop_mediated_treated_Estimate", "Prop_mediated_treated_P_value", "Prop_mediated_treated_Lower_CI_bound", "Prop_mediated_treated_Upper_CI_bound"] 
				
		print("Printing results to: %s" % csv_name)
		with open(csv_name, 'w', newline='') as csvfile:
			f = csv.writer(csvfile, quotechar='"', quoting=csv.QUOTE_NONNUMERIC, delimiter=',')
			f.writerow(header)
			for path_start in self.output:
				for outcome in self.output[path_start]:
					for mediator in self.output[path_start][outcome]:
						bk_rearranged = []
						med_rearranged = []
						p_list = self.output_p[path_start][outcome][mediator]
						if self.baron_and_kenny:
							bk_rearranged = [self.output_sobel[path_start][outcome][mediator][0], p_list[0], self.output_sobel[path_start][outcome][mediator][2], p_list[1],
											self.output_sobel[path_start][outcome][mediator][4], self.output_sobel[path_start][outcome][mediator][5], 
											self.output_sobel[path_start][outcome][mediator][6], self.output_sobel[path_start][outcome][mediator][7], p_list[2],
											self.output_sobel[path_start][outcome][mediator][9], self.output_sobel[path_start][outcome][mediator][10], p_list[3], 
											self.output_sobel[path_start][outcome][mediator][12], self.output_sobel[path_start][outcome][mediator][13], p_list[4]]
						if self.imai_and_keele: 
							med_rearranged = [self.output[path_start][outcome][mediator][0], p_list[5], 
												self.output[path_start][outcome][mediator][4], p_list[6], 
												self.output[path_start][outcome][mediator][8], p_list[7]] \
												 + self.output[path_start][outcome][mediator][2:4] +  self.output[path_start][outcome][mediator][6:8] +  self.output[path_start][outcome][mediator][10:12] \
												 + [self.output[path_start][outcome][mediator][12], p_list[8]] + self.output[path_start][outcome][mediator][14:16] \
												 + [self.output[path_start][outcome][mediator][16], p_list[9]] + self.output[path_start][outcome][mediator][18:20] \
												 + [self.output[path_start][outcome][mediator][20], p_list[10]] + self.output[path_start][outcome][mediator][22:24] \
												 + [self.output[path_start][outcome][mediator][24], p_list[11]] + self.output[path_start][outcome][mediator][26:28] \
												 + [self.output[path_start][outcome][mediator][28], p_list[12]] + self.output[path_start][outcome][mediator][30:32] \
												 + [self.output[path_start][outcome][mediator][32], p_list[13]] + self.output[path_start][outcome][mediator][34:36] \
												 + [self.output[path_start][outcome][mediator][36], p_list[14]] + self.output[path_start][outcome][mediator][38:] 
						f.writerow([path_start, outcome, mediator] + bk_rearranged + med_rearranged)

		
	def prep_data(self, path_start, mediator, outcome, controlDict=None, controlNames=None, zscoreRegression=None):
		"""Take dictionary data and return a Pandas DataFrame indexed by group_id.
		Column names are 'path_start', 'mediator' and 'outcome'
		"""
		ps_df = pd.DataFrame(data=list(collections.OrderedDict(sorted(path_start.items())).items()), columns=['group_id', 'path_start']).set_index(['group_id']).fillna(0)
		m_df = pd.DataFrame(data=list(mediator.items()), columns=['group_id', 'mediator']).set_index(['group_id'])
		o_df = pd.DataFrame(data=list(outcome.items()), columns=['group_id', 'outcome']).set_index(['group_id'])

		data = ps_df.join(m_df)
		data = data.join(o_df)
		if controlDict and controlNames: 
			for control in controlNames:
				data = data.join(pd.DataFrame(data=list(controlDict[control].items()), columns=['group_id', control]).set_index(['group_id']))
		data = data.dropna()
		return data

	def get_data(self, switch, outcome_field, location, features):
		"""Get data from outcomeGetter / featureGetter
		"""
		data = None
		if switch == "feat_as_path_start" and location == "path_start":
				data = features[outcome_field]
		elif switch == "feat_as_outcome" and location == "outcome":
				data = features[outcome_field]
		elif switch == "feat_as_control" and location == "control":
				data = features[outcome_field]
		elif switch == "default" and location == "mediator":
				data = features[outcome_field]
		else:
			data = dict((x, y) for x, y in self.outcomeGetter.getGroupAndOutcomeValues(outcomeField = outcome_field))
		return data

	def mediate(self, switch="default", p_correction_method = 'BH', 
				zscoreRegression = True, logisticReg = False):
		""" Runs the medition. 

		Parameters
		----------
		switch : str
			controls source (FeatureGetter or OutcomeGetter) of variables (path_starts, mediators, outcomes, controls)
		p_correction_method : str
			Name of p correction method
		zscoreRegression : boolean
			True if data is z-scored
		logisticReg : boolean
			True if running logistic regression
		
		Notes
		-----
		Data sources according to 'switch':
			"default": 
				FeatureGetter: mediators
				OutcomeGetter: path_starts and outcomes
			"feat_as_path_start":
				FeatureGetter: path_starts
				OutcomeGetter: mediators, outcomes and controls
			"feat_as_outcome":
				FeatureGetter: outcomes
				OutcomeGetter: path_starts, mediators and controls
			"feat_as_control":
				FeatureGetter: controls
				OutcomeGetter: path_starts, mediators and outcomes
			"no_features":
				OutcomeGetter: path_starts, mediators, outcomes and controls
		"""

		if "no_features" not in switch:
			(groups, allOutcomes, controls) = self.outcomeGetter.getGroupsAndOutcomes()

		if self.featureGetter:
			(allFeatures, featureNames) = self.featureGetter.getGroupNormsWithZerosFeatsFirst(groups=groups)

		if switch == "feat_as_path_start":
			if len(self.pathStartNames) == 0:
				self.pathStartNames = list(allFeatures.keys())
			numMeds = len(self.pathStartNames)

		elif switch == "feat_as_outcome":
			if len(self.outcomeNames) == 0:
				self.outcomeNames = list(allFeatures.keys())
			numMeds = len(self.outcomeNames)

		elif switch == "feat_as_control":
			if len(self.controlNames) == 0:
				self.controlNames = list(allFeatures.keys())
			numMeds = len(self.pathStartNames)

		elif switch == "no_features":
			numMeds = len(self.pathStartNames)

		elif switch == "default":
			if len(self.mediatorNames) == 0:
				self.mediatorNames = list(allFeatures.keys())
			numMeds = len(self.mediatorNames)
			
		mediation_count = 0
		total_mediations = str(len(self.pathStartNames)*len(self.mediatorNames)*len(self.outcomeNames))

		for path_start in self.pathStartNames:
			self.output[path_start] = {}
			self.output_sobel[path_start] = {}
			self.output_p[path_start] = {}

			for outcome in self.outcomeNames:
				self.output[path_start][outcome] = {}
				self.output_sobel[path_start][outcome] = {}
				self.output_p[path_start][outcome] = {}

				for mediator in self.mediatorNames:
					mediation_count += 1
					self.output_p[path_start][outcome][mediator] = []
					self.output[path_start][outcome][mediator] = []
					
					if len(self.controlNames) > 0:
						data = self.prep_data(self.get_data(switch, path_start, "path_start", allFeatures), 
							self.get_data(switch, mediator, "mediator", allFeatures), 
							self.get_data(switch, outcome, "outcome", allFeatures), 
							controlsDict, self.controlNames, zscoreRegression=zscoreRegression)
						
						control_formula = " + " + " + ".join(self.controlNames)
					else: 
						data = self.prep_data(self.get_data(switch, path_start, "path_start", allFeatures), 
							self.get_data(switch, mediator, "mediator", allFeatures), 
							self.get_data(switch, outcome, "outcome", allFeatures), 
							zscoreRegression=zscoreRegression)
						
						control_formula = ""
					
					outcome_exog = patsy.dmatrix("mediator + path_start " + control_formula + " -1", data,
													  return_type='dataframe')
					mediator_exog = patsy.dmatrix("path_start " + control_formula + " -1", data,
											 return_type='dataframe')
					direct_exog = patsy.dmatrix("path_start " + control_formula + " -1", data,
											 return_type='dataframe')

					if zscoreRegression:
						outcome_array = np.asarray(data['outcome']) if logisticReg else np.asarray(zscore(data['outcome']))
						mediator_array = np.asarray(data['mediator']) if logisticReg else np.asarray(zscore(data['mediator']))
						outcome_exog['mediator'] = zscore(outcome_exog['mediator'])
						outcome_exog['path_start'] = zscore(outcome_exog['path_start'])
						mediator_exog['path_start'] = zscore(mediator_exog['path_start'])
						direct_exog['path_start'] = zscore(direct_exog['path_start'])
					else:
						outcome_array = np.asarray(data['outcome'])
						mediator_array = np.asarray(data['mediator'])

					outcome_model = sm.OLS(outcome_array, outcome_exog)
					mediator_model = sm.OLS(mediator_array, mediator_exog)
					direct_model = sm.OLS(outcome_array, direct_exog)

					# classic mediation with Sobel Test
					if self.baron_and_kenny:
						outcome_results = outcome_model.fit()
						mediator_results = mediator_model.fit()
						direct_results = direct_model.fit()

						c = direct_results.params.get('path_start')
						c_prime = outcome_results.params.get('path_start')
						alpha = mediator_results.params.get('path_start')
						alpha_error = mediator_results.bse.get('path_start')
						beta = outcome_results.params.get('mediator')
						beta_error = outcome_results.bse.get('mediator')
						sobel_SE = sqrt(beta*beta*alpha_error*alpha_error + alpha*alpha*beta_error*beta_error)
						sobel = (alpha*beta)/ sobel_SE

						if p_correction_method.startswith("bonf"):
							c_p = direct_results.pvalues.get('path_start')*numMeds
							c_prime_p = outcome_results.pvalues.get('path_start')*numMeds
							alpha_p = mediator_results.pvalues.get('path_start')*numMeds
							beta_p = outcome_results.pvalues.get('mediator')*numMeds
							sobel_p = st.norm.sf(abs(sobel))*2*numMeds
						else:
							c_p = direct_results.pvalues.get('path_start')
							c_prime_p = outcome_results.pvalues.get('path_start')
							alpha_p = mediator_results.pvalues.get('path_start')
							beta_p = outcome_results.pvalues.get('mediator')
							sobel_p = st.norm.sf(abs(sobel))*2

						self.output_sobel[path_start][outcome][mediator] = np.array([c, c_p, c_prime, c_prime_p, c-c_prime, alpha*beta, alpha, alpha_error, alpha_p, beta, beta_error, beta_p, sobel, sobel_SE, sobel_p])
						self.output_p[path_start][outcome][mediator] = self.output_p[path_start][outcome][mediator] + [c_p, c_prime_p, alpha_p, beta_p, sobel_p ]

					# imai_and_keele mediation method
					if self.imai_and_keele:
						tx_pos = [outcome_exog.columns.tolist().index("path_start"),
								  mediator_exog.columns.tolist().index("path_start")]
						med_pos = outcome_exog.columns.tolist().index("mediator")

						med = Mediation(outcome_model, mediator_model, tx_pos, med_pos)

						med_result = med.fit(method=self.mediation_method, n_rep=self.boot_number)
						summary = med_result.summary(p_correction_method = p_correction_method, numMeds=numMeds)
						summary_array = np.reshape(summary.values, 40).tolist()
						
						self.output[path_start][outcome][mediator] = summary_array
						self.output_p[path_start][outcome][mediator] = self.output_p[path_start][outcome][mediator] + summary["P-value"].tolist()

					print("Mediation number " + str(mediation_count) + " out of " + total_mediations)
					
					if len(self.controlNames) > 0:
						print("Path Start: %s, Mediator: %s, Outcome: %s, Controls: %s" % (path_start, mediator, outcome, ", ".join(self.controlNames)))
					else:
						print("Path Start: %s, Mediator: %s, Outcome: %s" % (path_start, mediator, outcome))
					if self.baron_and_kenny:
						print("C: %s, C_p: %s, C': %s, C'_p: %s" % (str(c), str(c_p), str(c_prime), str(c_prime_p)))
						print("C-C': %s, alpha*beta: %s" % (str(c-c_prime), str(alpha*beta)))
						print("alpha: %s, alpha_error: %s, alpha_p: %s" % (str(alpha), str(alpha_error), str(alpha_p)))
						print("beta: %s, beta_error: %s, beta_p: %s" % (str(beta), str(beta_error), str(beta_p)))
						print("Sobel z-score: %s, Sobel SE: %s, Sobel p: %s" % (str(sobel), str(sobel_SE), str(sobel_p)))
					if self.imai_and_keele:
						print(summary)
					print("")

		if p_correction_method and not p_correction_method.startswith("bonf"):
			p_list = []
			if self.baron_and_kenny:
				p_list = p_list + ["C_p", "C'_p", "alpha_p", "beta_p", "sobel_p"]
			if self.imai_and_keele:
				p_list = p_list + ["Prop_mediated_average_P_value", "ACME_average_P_value", "ADE_average_P_value", "ACME_treated_P_value", 
					"ACME_control_P_value", "ADE_treated_P_value", "ADE_control_P_value", "Total_effect_P_value", 
					"Prop_mediated_control_P_value", "Prop_mediated_treated_P_value"]
			p_dict = dict()
			r_dict = dict()
			for path_start in self.pathStartNames:
				for outcome in self.outcomeNames:
					for i, p in enumerate(p_list):
						p_dict[p] = dict()
						for mediator in self.mediatorNames:
							p_dict[p][mediator] = self.output_p[path_start][outcome][mediator][i]
						p_dict[p] = pCorrection(p_dict[p], p_correction_method, [0.05, 0.01, 0.001], rDict = None)
					for mediator in self.mediatorNames:
						self.output_p[path_start][outcome][mediator] = [p_dict[p][mediator] for p in p_list]




###############################
## correlates results between two csvs of results
## csvs in arguments:
## test-retest.rmatrix.py <rmatrix_csv> <rmatrix_csv>

import csv
import sys
from scipy.stats.stats import pearsonr, spearmanr
from pprint import pprint
from numpy import array, concatenate, log, sqrt, isnan, isinf,  std, clip, nan_to_num

#csv.quotechar = '\x07'
csv.field_size_limit(sys.maxsize)
class DDLA:

    ignoreColumns = set(['p', 'N', 'freq', 'feature'])
    file1 = None
    file2 = None
    data = None
    header = None
    outputData = None
    
    def __init__(self, file1, file2, outputFile = None):
        self.file1 = file1
        self.file2 = file2
        self.outputFile = outputFile if outputFile else '-'.join([file1[:-4], file2[:-4]])+'.csv'
        self.data = list()
        self.header = list()
    
    def signed_r_square(self,r):
        r2 = float(r)**2
        if r < 0: 
            return -1 * r2
        return r2

    def signed_r_log(self,r):
        if r < 0: 
            return -1 * log((-1*float(r))+1)
        return log(r+1)

    def compare_correl(self,ra, na, rb, nb):
        if any(not isnan(i) for i in [ra, na, rb, nb]):
            (ra, rb) = min(1, max(-1, ra)), min(1, max(-1, rb))
            raplus = 1*ra+1
            raminus = 1-ra
            rbplus = 1*rb+1
            rbminus = 1-rb

            za = (log(raplus)-log(raminus))/2
            zb = (log(rbplus)-log(rbminus))/2

            se = sqrt((1/(na-3))+(1/(nb-3)))
            z  = (za-zb)/se

            z2 = abs(z)

            p2 = (((((.000005383*z2+.0000488906)*z2+.0000380036)*z2+.0032776263)*z2+.0211410061)*z2+.049867347)*z2+1

            p2 = pow(p2, -16)

            return p2

        else:
            return float('nan')

    def write2CSV(self, dataDict, features):
        toWrite = dataDict['data'].tolist()
        for i in range(len(features)): toWrite[i].insert(0,features[i])
        toWrite.sort(key=lambda x: x[0])
        self.outputData = toWrite
        with open(self.outputFile,'w+') as csv_file:
            write = csv.writer(csv_file)
            write.writerow(dataDict['header'])
            write.writerows(toWrite)

    def add2Output(self,csvOutput, outcome_data, outcome_name, featsInOrder):
        data = self.data
        ordered = ['value', 'p', 'freq', 'N']
        originalData0 = array([[data[0][outcome_name][feat][col] for col in ordered]
                               for feat in featsInOrder])
        originalData1 = array([[data[1][outcome_name][feat][col] for col in ordered]
                               for feat in featsInOrder])
        originalData = concatenate((originalData0,originalData1), axis=1)
        csvOutput['header'].extend([outcome_name, 'p'] + ['r_0', 'p_0', 'freq_0', 'N_0'] + ['r_1', 'p_1', 'freq_1', 'N_1'])
        self.header = csvOutput['header']

        if 'data' in list(csvOutput.keys()):
            csvOutput['data'] = concatenate((csvOutput['data'], nan_to_num(outcome_data)), axis = 1)
        else:
            csvOutput['data'] = nan_to_num(outcome_data)
        csvOutput['data'] = concatenate((csvOutput['data'], originalData),axis=1)

    def get_next(self,some_iterable, window=1):
        from itertools import tee, islice, zip_longest
        items, nexts = tee(some_iterable, 2)
        nexts = islice(nexts, window, None)
        return zip_longest(items, nexts)

    def print_sorted(self, dataDict, features):
        data = dataDict['data'].tolist()
        for i in range(len(features)): data[i].insert(0,features[i])
        data.sort(key=lambda x: -x[1])

    def load_data(self):
        data = self.data
        fins = [self.file1, self.file2]
        for split in range(len(fins)):
            fin = open(fins[split], 'r')
            reader = csv.reader(fin)
            data.append(dict())

            #get headers:
            headers = None
            while True:
                headers = next(reader)
                if len(headers) > 1 and headers[0][:6] != 'Namesp':
                    #remove blanks and namespace lines
                    break

            #setup data dict, using column names
            for h in headers:
                h = h.strip()
                if h not in (self.ignoreColumns):
                    data[split][h] = dict()

            #read all rows:
            for row in reader:
                if row and row[0] == 'SORTED:':
                    print("found SORTED:", row)
                    break
                if len(row) > 1:
                    feat = row[0].strip()
                    column_used = None
                    for i in range(1, len(row)):
                        # Going through the entire row, entry by entry
                        if i < len(headers):
                            column_name = None
                            if headers[i] in data[split]:
                                # New Outcome
                                column_used = headers[i]
                                data[split][column_used][feat] = dict()
                                column_name = 'value'
                            else:
                                column_name = headers[i]
                            cell_value = row[i].strip()
                            if cell_value :#and not isnan(float(cell_value)):
                                data[split][column_used][feat][column_name] = float(cell_value)

    def outputForTagclouds(self, sizeField = 1, colorField = "dr"):
        correls = dict()
        outcomes = self.header[1::10]
        correls = {outcome: dict() for outcome in outcomes}
        # print self.header[1::10] # color
        # print self.header[sizeField+2::10] # first country correlations = size
        # print self.header[sizeField+6::10] # secnd country correlations = size
        for row in self.outputData:
            drs = row[1::10]
            rs = row[sizeField+2::10]
            ps = row[sizeField+3::10]
            Ns = row[sizeField+5::10]
            for i,o in enumerate(outcomes):
                correls[o][row[0]] = (rs[i], ps[i], Ns[i], drs[i])
        return correls

    def differential(self):
        data = self.data
        self.load_data()
        #from pprint import pprint
        #pprint([(k, v) for k,v in data[1]['is_student'].iteritems() if k == 'den'][:10])
        commonOutcomes = set(data[0].keys()) & set(data[1].keys()) 

        sumRs = float(0)
        sumR2s = float(0)
        sumRhos = float(0)
        sumRho2s = float(0)
        csvOutput = {'header': ['feature',]}
        commonFeats = None

        for outcome in sorted(commonOutcomes):
            print("\n%s\n%s" % (outcome, '='*len(outcome)))
            feats0 = set(data[0][outcome].keys())
            feats1 = set(data[1][outcome].keys())
            commonFeats = feats0 & feats1
            print("Number of feats in first results:  %d" % len(feats0))
            print("Number of feats in second results: %d" % len(feats1))
            print("Number of feats in common:         %d" % len(commonFeats))

            # Getting data in the right format
            commonFeats = list(commonFeats)
            list0 = nan_to_num(array([data[0][outcome][feat]['value'] for feat in commonFeats]))
            list1 = nan_to_num(array([data[1][outcome][feat]['value'] for feat in commonFeats]))
            list0_n = array([data[0][outcome][feat]['N'] for feat in commonFeats])
            list1_n = array([data[1][outcome][feat]['N'] for feat in commonFeats])

            # Comparing individual correlations
            diffs = list0 - list1
            output = array([[diffs[i], self.compare_correl(list0[i], list0_n[i], list1[i], list1_n[i])] for i in range(len(list0))])
            self.add2Output(csvOutput, output, outcome, commonFeats)

            sorted_pairs = sorted(zip(diffs, commonFeats), key= lambda x_y: -x_y[0])
            sorted_words = [x_y1[1] for x_y1 in sorted_pairs]
            print('Decreasing differences: <top 10 most correlated with %s> ... <top 10 most correlated with %s>' % (self.file1[:-4], self.file2[:-4] ))
            print(', '.join(sorted_words[:10])+' ... '+', '.join(sorted_words[-10:]))


            # From previous script, comparing all the r's
            list0r2 = array([self.signed_r_log(r) for r in list0])
            list1r2 = array([self.signed_r_log(r) for r in list1])
            (r, p) = pearsonr(list0, list1)
            (r2, p2) = pearsonr(list0r2, list1r2)
            print("pearson r of rs:               %10.4f (%.6f)" % (r, p))
            print("pearson r of signed log(r)s:   %10.4f (%.6f)" % (r2, p2))
            sumRs += r
            sumR2s += r2
            (rho, p) = spearmanr(list0, list1)
            (rho2, p2) = spearmanr(list0r2, list1r2)
            print("spearman rho of rs:            %10.4f (%.6f)" % (rho, p))
            print("spearman rho of signed log(r)s:%10.4f (%.6f)" % (rho2, p2))
            sumRhos += rho
            sumRho2s += rho2

        self.outputData = csvOutput
        self.write2CSV(csvOutput, commonFeats)
        # print_sorted(csvOutput, commonFeats)


        print("\nAVERAGE RESULTS\n===============")
        print("pearson r of rs:               %10.4f " % (sumRs / float(len(commonOutcomes))))
        print("pearson r of log(r)s:          %10.4f " % (sumR2s / float(len(commonOutcomes))))
        print("spearman rho of rs:            %10.4f " % (sumRhos / float(len(commonOutcomes))))
        print("spearman rho of log(r)s:       %10.4f " % (sumRho2s / float(len(commonOutcomes))))

#### TODO: calculate number in common in top 100 (i.e. as if comparing word clouds)

if __name__ == '__main__':
    fins = sys.argv[1:3]
    ddla = DDLA(fins[0], fins[1])
    print(dir(ddla))
    ddla.differential()



"""
Module for defining constants and helper methods
"""

import sys, os, getpass
import re
from random import randint
from pathlib import Path

sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))

#math / stats:
from math import floor, log
from numpy import sqrt, log2, array, mean, std, isnan, fabs, round
from numpy.random import permutation
import numpy as np
from scipy.stats import zscore, norm, t
from scipy.sparse import csr_matrix
from sklearn.metrics import roc_auc_score
from sklearn.linear_model import LogisticRegression
import statsmodels.stats.multitest as mt

#DB INFO:
USER = getpass.getuser()

MAX_ATTEMPTS = 5 #max number of times to try a query before exiting
PROGRESS_AFTER_ROWS = 5000 #the number of rows to process between each progress updated
FEATURE_TABLE_PREFIX = 'feats_'
MYSQL_ERROR_SLEEP = 4 #number of seconds to wait before trying a query again (incase there was a server restart
MYSQL_BATCH_INSERT_SIZE = 10000 # how many rows are inserted into mysql at a time
MAX_SQL_SELECT = 1000000 # how many rows are selected at a time
MYSQL_HOST = '127.0.0.1'
VARCHAR_WORD_LENGTH = 36 #length to allocate var chars per words
LOWERCASE_ONLY = True #if the db is case insensitive, set to True
MAX_TO_DISABLE_KEYS = 100000 #number of groups * n must be less than this to disable keys
MAX_SQL_PRINT_CHARS = 256

##Corpus Settings:
DEF_CORPDB = 'dla_tutorial'
DEF_CORPTABLE = ''
DEF_CORREL_FIELD = 'user_id'
DEF_MESSAGE_FIELD = 'message'
DEF_MESSAGEID_FIELD = 'message_id'
DEF_ENCODING = 'utf8mb4'
DEF_UNICODE_SWITCH = True
DEF_LEXTABLE = 'wn_O'
DEF_DATE_FIELD = 'updated_time'
DEF_COLLATIONS = {
        'utf8mb4': 'utf8mb4_bin',
        'utf8': 'utf8_general_ci',
        'latin1': 'latin1_swedish_ci',
        'latin2': 'latin2_general_ci',
        'ascii': 'ascii_general_ci',
    }
DEF_MYSQL_ENGINE = 'MYISAM'

##lexInterface settings
DEF_TERM_FIELD = 'term'
DEF_MIN_WORD_FREQ = 1000
DEF_NUM_RAND_MESSAGES = 100
MAX_WRITE_RECORDS = 1000 #maximum number of records to write at a time (for add_terms...)

##Outcome settings
DEF_OUTCOME_TABLE = ''
DEF_OUTCOME_FIELD = ''
DEF_OUTCOME_FIELDS = []
DEF_OUTCOME_CONTROLS = []
DEF_GROUP_FREQ_THRESHOLD = 1000 #min. number of total feature values that the group has, to use it
DEF_SHOW_FEAT_FREQS = True
DEF_MAX_TC_WORDS = 100
DEF_MAX_TOP_TC_WORDS = 15
DEF_TC_FILTER = True
DEF_WEIGHTS = ''
DEF_LOW_VARIANCE_THRESHOLD = 0.0

##Feature Settings:
DEF_N = int(1)
DEF_FEAT_NAMES = ['honor']
DEF_MIN_FREQ = int(1) #min frequency per group to keep (don't advise above 1)
DEF_P_OCC = float(.01) #percentage of groups a feature must appear in, to keep it
DEF_PMI = 3.0
DEF_MIN_FEAT_SUM = 0 #minimum sum of feature total to keep
DEF_LEXICON_DB = 'dlatk_lexica'
DEF_FEAT_TABLE = 'feat$1gram$messages_en$user_id$16to16$0_01'
DEF_COLLOCTABLE = 'test_collocs'
DEF_COLUMN_COLLOC = "feat"
DEF_COLUMN_PMI_FILTER = "pmi_filter_val"
DEF_P = 0.05 # p value for printing tagclouds
DEF_P_CORR = 'BH' #Benjamini, Hochberg
DEF_P_MAPPING = { # maps old R method names to statsmodel names
        "holm": "holm",
        "hochberg": "simes-hochberg",
        "simes": "simes",
        "hommel": "hommel",
        "bonferroni": "bonferroni",
        "bonf": "bonferroni",
        "BH": "fdr_bh",
        "BY": "fdr_by",
        "fdr": "fdr_bh",
        "sidak": "sidak",
        "holm-sidak": "holm-sidak",
        "simes-hochberg": "simes-hochberg",
        "fdr_bh": "fdr_bh",
        "fdr_by": "fdr_by",
        "fdr_tsbh": "fdr_tsbh",
        "fdr_tsbky": "fdr_tsbky",
    }
DEF_CONF_INT = 0.95
DEF_TOP_MESSAGES = 10

DEF_BERT_MODEL='base-uncased'
DEF_BERT_AGGREGATION=['mean']
DEF_BERT_LAYER_AGGREGATION=['concatenate']
DEF_BERT_LAYERS=[10]
DEF_TRANS_WORD_AGGREGATION = ['mean']


##Prediction Settings:
DEF_MODEL = 'ridgecv'
DEF_CLASS_MODEL = 'svc'
DEF_COMB_MODELS = ['ridgecv']
DEF_FOLDS = 5
DEF_OUTLIER_THRESHOLD = 2.5
DEF_RP_FEATURE_SELECTION_MAPPING = {
    'magic_sauce': 'Pipeline([("1_mean_value_filter", OccurrenceThreshold(threshold=int(sqrt(X.shape[0]*10000)))), ("2_univariate_select", SelectFwe(f_regression, alpha=60.0)), ("3_rpca", PCA(n_components=max(int(X.shape[0]/(len(self.featureGetters)+0.1)), min(min(50,X.shape[0]), X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',

    'magic_sauce_light': 'Pipeline([("1_mean_value_filter", OccurrenceThreshold(threshold=int(sqrt(X.shape[0]*10000)))), ("2_univariate_select", SelectFwe(f_regression, alpha=100.0)), ("3_rpca", PCA(n_components=max(int(X.shape[0]/(len(self.featureGetters)+0.3)), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',

    'magic200': 'Pipeline([("1_univariate_select", SelectFwe(alpha=60.0, score_func=f_regression)), ("2_rpca", PCA(copy=True, iterated_power=3, n_components=200, random_state=42, whiten=False, svd_solver="randomized"))])',

    'topic_ngram_ms': 'Pipeline([("1_mean_value_filter", OccurrenceThreshold(threshold=int(sqrt(X.shape[0]*10000)))), ("2_univariate_select", SelectFwe(f_regression, alpha=60.0)), ("3_rpca", PCA(n_components=max(int(2*(X.shape[0]*.20)/len(self.featureGetters) if X.shape[1] > 2000 else 2*(X.shape[0]*.75)/len(self.featureGetters)), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',

    'magic_sauce_1pct': 'Pipeline([("1_mean_value_filter", OccurrenceThreshold(threshold=int(sqrt(X.shape[0]*10000)))), ("2_univariate_select", SelectFwe(f_regression, alpha=60.0)), ("3_rpca", PCA(n_components=min(int((X.shape[0]/(len(self.featureGetters)+0.1))*.001), int(X.shape[1]*0.2)), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',

    'topic_ngram_ms': 'Pipeline([("1_mean_value_filter", OccurrenceThreshold(threshold=int(sqrt(X.shape[0]*10000)))), ("2_univariate_select", SelectFwe(f_regression, alpha=60.0)), ("3_rpca", PCA(n_components=max(int(2*(X.shape[0]*.20)/len(self.featureGetters) if X.shape[1] > 2000 else 2*(X.shape[0]*.75)/len(self.featureGetters)), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',

    'ms_old1':'Pipeline([("1_univariate_select", SelectFwe(f_regression, alpha=0.60)), ("2_rpca", RandomizedPCA(n_components=max(min(int(X.shape[1]*.10), int(X.shape[0]/len(self.featureGetters))), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3))])',

    'ms_old2': 'Pipeline([("1_univariate_select", SelectFwe(f_regression, alpha=60.0)), ("2_rpca", RandomizedPCA(n_components=max(min(int(X.shape[1]*.10), int(X.shape[0]/max(1.5,len(self.featureGetters)))), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3))])', 

    'univariatefwe': 'SelectFwe(f_regression, alpha=60.0)',

    'pca': 'PCA(n_components=max(min(int(X.shape[1]*.5), int(X.shape[0]/max(1.5,len(self.featureGetters)))), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized")',
    'none': None,
}
DEF_CP_FEATURE_SELECTION_MAPPING = {
    'magic_sauce': 'Pipeline([("1_univariate_select", SelectFwe(f_classif, alpha=30.0)), ("2_rpca", PCA(n_components=min(max(min(int(X.shape[1]*.10), int(X.shape[0]/max(1.5,len(self.featureGetters)))), 50), X.shape[1]), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized"))])',
    'univariatefwe': 'SelectFwe(f_classif, alpha=30.0)',
    'univariatefwe30': 'SelectFwe(f_classif, alpha=30.0)',
    'univariatefwe10': 'SelectFwe(f_classif, alpha=10.0)',
    'univariatefwe60': 'SelectFwe(f_classif, alpha=60.0)',
    'pca': 'PCA(n_components=max(min(int(X.shape[1]*.10), int(X.shape[0]/max(1.5,len(self.featureGetters)))), min(50, X.shape[1])), random_state=42, whiten=False, iterated_power=3, svd_solver="randomized")',
    'none': None,
}
DEFAULT_MAX_PREDICT_AT_A_TIME = 100000
DEFAULT_RANDOM_SEED = 42

##Mediation Settings:
DEF_MEDIATION_BOOTSTRAP = 1000
DEF_OUTCOME_PATH_STARTS = []
DEF_OUTCOME_MEDIATORS = []
DEF_MAX_MED_SUMMARY_SIZE = 10 # maximum number of results to print in summary for each path start / outcome pair

##LDA settings
DEF_LDA_MSG_TABLE = 'messages_en_lda$msgs_en_tok_a30'

##Language filtering settings
AVAILABLE_LANGUAGES = """
Available:  af, am, an, ar, as, az, be, bg, bn, br,
bs, ca, cs, cy, da, de, dz, el, en, eo, es, et, eu,
fa, fi, fo, fr, ga, gl, gu, he, hi, hr, ht, hu, hy,
id, is, it, ja, jv, ka, kk, km, kn, ko, ku, ky, la,
lb, lo, lt, lv, mg, mk, ml, mn, mr, ms, mt, nb, ne,
nl, nn, no, oc, or, pa, pl, ps, pt, qu, ro, ru, rw,
se, si, sk, sl, sq, sr, sv, sw, ta, te, th, tl, tr,
ug, uk, ur, vi, vo, wa, xh, zh, zu"""
DEF_LANG_FILTER_CONF = .80
DEF_SPAM_FILTER = 0.2 # threshold for removing spam users

## Other tools
DEF_TOOLS_PATH = str(Path.home()) + '/dlatk_tools'

DEF_STANFORD_SEGMENTER = DEF_TOOLS_PATH + '/stanford-segmenter/segment.sh'
DEF_STANFORD_POS_MODEL = DEF_TOOLS_PATH + '/stanford-postagger/models/english-bidirectional-distsim.tagger'
DEF_STANFORD_PARSER = DEF_TOOLS_PATH + '/stanford-parser'

##CoreNLP settings
DEF_CORENLP_DIR = DEF_TOOLS_PATH + '/corenlp-python'
DEF_CORENLP_SERVER_COMMAND = './corenlp/corenlp.py'
DEF_CORENLP_PORT = 20202   #default: 20202
#CORE NLP PYTHON SERVER COMMAND (must be running): ./corenlp/corenlp.py -p 20202 -q

DEF_WORDCLOUD_JAR = DEF_TOOLS_PATH + "/wordcloud/ibm-word-cloud.jar"

TAG_RE = re.compile(r'<[^>]+>')
URL_RE = re.compile(r'(?:http[s]?\:\/\/)?(?:[\w\_\-]+\.)+(?:com|net|gov|edu|info|org|ly|be|gl|co|gs|pr|me|cc|us|uk|gd|nl|ws|am|im|fm|kr|to|jp|sg|int|mil|arpa|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|bq|br|bs|bt|bv|bw|by|bz|bzh|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cw|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zr|zw)+(?:\/[^\s ]+)?')

DEF_TOPIC_LEX_METHOD = 'csv_lik'

PERMA_SORTED = ['P+',
                'P-',
                'E+',
                'E-',
                'R+',
                'R-',
                'M+',
                'M-',
                'A+',
                'A-',
                'ope',
                'con',
                'ext',
                'agr',
                'neu'
                ]

DEF_CENSOR_DICT = {"fuck":"f**k",
        "pussy":"p**sy",
        "bitch":"b**ch",
        "bitches":"b**ches",
        "fuckn":"f**kn",
        "motherfucker":"motherf**ker",
        "fucked":"f**ked",
        "fucking":"f**king",
        "dick":"d**k",
        "dickhead":"d**khead",
        "cock":"c**k",
        "shit":"s**t",
        "bullshit":"bulls**t",
        "fuckin":"f**kin",
        "whore":"w**re",
        "hoe":"h**",
        "hoes":"h**s",
        "nigga":"n**ga",
        "niggas":"n**gas",
        "nigga's":"n**ga's",
        "niggaz":"n**gaz"}
GIGS_OF_MEMORY = 512 #used to determine when to use queries that hold data in memory
CORES = 32 #used to determine multi-processing

POSSIBLE_VALUE_FUNCS = [
    lambda d: 1,
    lambda d: float(d),
    lambda d: sqrt(float(d)),
    lambda d: log(float(d) + 1),
    lambda d: log2(float(d) + 1),
    lambda d: 2*sqrt(d+3/float(8))
]

##Meta settings
DEF_INIT_FILE = 'initFile.ini'
WARNING_STRING = "\n".join(["#"*68, "#"*68, "WARNING: %s", "#"*68, "#"*68])

##########################################################
### Static Module Methods
##
#

def alignDictsAsLists(dict1, dict2):
    """Converts two dictionaries to vectors, where the values are aligned by keys"""
    coFields = frozenset(list(dict1.keys())).intersection(frozenset(list(dict2.keys())))
    list1 = [dict1[c] for c in coFields]
    #list1 = [dict1[c] for c in coFields] #equivalent to above
    list2 = [dict2[c] for c in coFields]
    return (list1, list2)

def alignDictsAsX(X, sparse=False, returnKeyList=False):
    """turns a list of dicts for into a matrix X"""
    keys = frozenset()
    if sparse:
        keys = frozenset([item for sublist in X for item in sublist])  # union X keys
    else:
        keys = frozenset(list(X[0].keys())). intersection(*[list(x.keys()) for x in X[1:]])  # intersect X keys
    keys = list(keys)  # to make sure it stays in order
    if sparse:
        keyToIndex = dict([(keys[i], i) for i in range(len(keys))])
        row = []
        col = []
        data = []
        for c in range(len(X)):
            column = X[c]
            for keyid, value in column.items():
                if keyid in keyToIndex:
                    row.append(keyToIndex[keyid])
                    col.append(c)
                    data.append(value)
        sparseX = csr_matrix((data, (row, col)))
        if returnKeyList:
            return (sparseX, keys)
        else:
            return (sparseX)
    else:
        listX = [[x[k] for x in X] for k in keys]
        if returnKeyList:
            return (listX, keys)
        else:
            return (listX)

def alignDictsAsXy(X, y):
    """turns a list of dicts for x and a dict for y into a matrix X and vector y"""
    keys = frozenset(list(y.keys()))
    keys = keys.intersection(*[list(x.keys()) for x in X])
    keys = list(keys) #to make sure it stays in order
    listy = [y[k] for k in keys]
    # Order of columns of X is preserved
    listX = [[x[k] for x in X] for k in keys]
    # print type(listy), type(listy[0])
    import decimal
    listy = [float(v) for v in listy] if isinstance(listy,list) and isinstance(listy[0],decimal.Decimal) else listy
    # print type(listy), type(listy[0])
    return (listX, listy)

def bonfPCorrection(tup, numFeats):
    """given tuple with second entry a p value, multiply by num of feats and return"""
    return (tup[0], tup[1]*numFeats) + tup[2:]

def cohensD(x, y):
    """Calculate Cohen's D effect size

    Parameters
    ----------
    x : :obj:`list`
        list of observations
    y : :obj:`list`
        list of class (0 or 1) values for each observation
    """
    x0, x1 = list(), list()
    for i, e in enumerate(y):
        if e > 0:
            x1.append(x[i])
        else:
            x0.append(x[i])
    return (mean(x1) - mean(x0)) / (sqrt((std(x0, ddof=1) ** 2 + std(x1, ddof=1) ** 2) / 2))


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i+n]

def conf_interval(r, samp_size, percent=DEF_CONF_INT):
    """calculates two sided confidence interval for pearson correlation coefficient (r)"""
    tup = ()
    if not np.isnan(r):
        z = np.arctanh(r)
        sigma = (1/((samp_size-3)**0.5))
        cint = z + np.array([-1, 1]) * sigma * norm.ppf((1+percent)/2)
        tup = tuple(np.tanh(cint))
    else:
        tup = (np.nan, np.nan)
    return tup

def meanXperY(x, y):
    """find the mean x for each label y"""
    sums = dict()
    counts = dict()
    for i in range(len(x)):
        try:
            sums[y[i]] += x[i]
            counts[y[i]] += 1
        except KeyError:
            sums[y[i]] = x
            counts[y[i]] = 1
    means = dict()
    for yKey in sums:
        means[yKey] = sums[yKey] / float(counts[yKey])
    return means

def fiftyChecks(args):
    Xc, Xend, y, check = args
    np.random.seed()
    lr = LogisticRegression(penalty='l2', C=1000000, fit_intercept=True)
    if Xc is not None:
        r = sum([roc_auc_score(y, lr.fit(newX,y).predict_proba(newX)[:,1]) > check for newX in [np.append(Xc, permutation(Xend), 1) for i in range(50)]])
    else:
        r = sum([roc_auc_score(y, lr.fit(newX,y).predict_proba(newX)[:,1]) > check for newX in [permutation(Xend).reshape(len(Xend),1) for i in range(50)]])
    return r

def getGroupFreqThresh(correl_field=None):
    """set group_freq_thresh based on level of analysis"""
    group_freq_thresh = DEF_GROUP_FREQ_THRESHOLD
    if correl_field:
        if any(field in correl_field.lower() for field in ["mess", "msg"]) or correl_field.lower().startswith("id"):
            group_freq_thresh = 1
        elif any(field in correl_field.lower() for field in ["user", "usr", "auth"]):
            group_freq_thresh = 500
        elif any(field in correl_field.lower() for field in ["county", "cnty", "cty", "fips"]):
            group_freq_thresh = 40000
    return group_freq_thresh

def getMetric(logistic=False, cohensd=False, idp=False, spearman=False, controls=False):
    """Return a string which best represents the metric used in the analysis. String is used in output (csv, wordcloud, etc.)"""
    if logistic:
        metric = 'B'
    elif cohensd:
        metric = 'D'
    elif idp:
        metric = 'R'
    elif spearman:
        metric = 'rho'
    elif controls:
        metric = 'beta'
    else:
        metric = 'r'
    return metric

def permaSortedKey(s):
    if isinstance(s, (list, tuple)):
        s = str(s[0])
    if isinstance (s, (float, int)):
        return s
    return s

def pCorrection(pDict, method=DEF_P_CORR, pLevelsSimes=[0.05, 0.01, 0.001], rDict = None):
    """returns corrected p-values given a dict of [key]=> p-value"""
    method = DEF_P_MAPPING[method]
    pLevelsSimes = list(pLevelsSimes) #copy so it can be popped
    new_pDict = {}
    if method == 'simes':
        n = len(pDict)

        #pDictTuples = [[k, 1 if isnan(float(v)) else v] for k, v in pDict.iteritems()]
        pDictTuples = [[k, v] for k, v in pDict.items()]
        sortDict = rDict if rDict else pDict

        sortedPTuples = sorted(pDictTuples, key=lambda tup: 0 if isnan(sortDict[tup[0]]) else fabs(sortDict[tup[0]]), reverse=True if rDict else False)
        ii = 0
        rejectRest = False
        pMax = pLevelsSimes.pop()
        for ii in range(len(sortedPTuples)):
            if rejectRest:
                sortedPTuples[ii][1] = 1
            else:
                newPValue = sortedPTuples[ii][1] * n / (ii + 1)
                if newPValue < pMax:
                    sortedPTuples[ii][1] = round((pMax - .00001) * pMax, 5)
                else:
                    while len(pLevelsSimes) > 0 and newPValue >= pMax:
                        pMax = pLevelsSimes.pop()
                    if len(pLevelsSimes) == 0:
                        if newPValue < pMax:
                            sortedPTuples[ii][1] = round((pMax - .00001) * pMax, 5)
                        else:
                            rejectRest = True
                            sortedPTuples[ii][1] = 1
                    else:
                        sortedPTuples[ii][1] = round((pMax - .00001) * pMax, 5)
        new_pDict = dict(sortedPTuples)
    else:
        keys = list(pDict.keys())
        reject, pvals_corrected, alphacSidak, alphacBonf  = mt.multipletests(
            pvals=[1 if isnan(p) else p for p in [pDict[k] for k in keys]], method=method)
        i = 0
        for key in keys:
            new_pDict[key] = pvals_corrected[i]
            i+=1
    return new_pDict

def reverseDictDict(d):
    """reverses the order of keys in a dict of dicts"""
    assert isinstance(next(iter(d.values())), dict), 'reverseDictDict not given a dictionary of dictionaries'
    revDict = dict()
    for key1, subd in d.items():
        for key2, value in subd.items():
            if not key2 in revDict:
                revDict[key2] = dict()
            revDict[key2][key1] = value
    return revDict

def rgbColorMix(fromColor, toColor, resolution, randomness = False):
    #fromColor, toColor rgb (255 max) tuple
    #resolution, how many truple to return inbetween
    #(starts at from, but comes up one short of ToColor)
    (fromColor, toColor) = (array(fromColor), array(toColor))
    fromTo = toColor - fromColor #how to get from fromColor to toColor
    fromToInc = fromTo / float(resolution)
    gradientColors = []
    for i in range(resolution):
        gradientColors.append(tuple([int(x) for x in round(fromColor + (i * fromToInc))]))
    if randomness:
        for i in range(len(gradientColors)):
            color = gradientColors[i]
            newcolor = []
            for value in color:
                value += 20 - randint(0, 40)
                value = max(0, min(255, value))
                newcolor.append(value)
            gradientColors[i] = tuple(newcolor)

    return gradientColors

def rowsToColumns(X):
    """Changes each X from being represented in columns to rows """
    return [X[0:, c] for c in range(len(X[0]))]

def stratifiedZScoreybyX0(X, y):
    """zscores based on the means of all unique ys"""
    #TODO: probably faster in vector operations
    #first separate all rows for each y
    X0sToYs = dict()
    for i in range(len(y)):
        try:
            X0sToYs[X[i][0]].append(y[i])
        except KeyError:
            X0sToYs[X[i][0]] = [y[i]]

    #next figure out the mean for x, for each unique y
    meanYforX0s = []
    for ys in X0sToYs.values():
        meanYforX0s.append(mean(ys)) #should turn into a row

    #figure out mean and std-dev for meanXs:
    meanOfMeans = mean(meanYforX0s) #should be a row
    stdDevOfMeans = std(meanYforX0s) #shoudl be a row

    #apply to y:
    newY = []
    for yi in y:
        newY.append((yi - meanOfMeans)/float(stdDevOfMeans))

    return (zscore(X), newY)

def switchColumnsAndRows(X):
    """Toggles X between rows of columns and columns of rows """
    if not isinstance(X, np.ndarray): X = array(X)
    return array([X[0:, c] for c in range(len(X[0]))])

def tupleToStr(tp):
    """Takes in a tuple and returns a string with spaces instead of commas"""
    if isinstance(tp, tuple):
        return ';'.join([tupleToStr(t) for t in tp])
    return tp

def unionDictsMaxOnCollision(d1, d2):
    """unions d1 and 2 always choosing the max between the two on a collision"""
    newD = d1.copy()
    for k2, v2 in d2.items():
        if (not k2 in newD) or v2 > newD[k2]:
            newD[k2] = v2
    return newD

def warn(string, attention=False):
    if attention: string = WARNING_STRING % string
    print(string, file=sys.stderr)

#Local maintenance:
def getReportingInt(reporting_percent, iterable_length):
    return max(1, floor(reporting_percent * iterable_length))

def report(object_description, ii, reporting_int, iterable_length):
    if ii % reporting_int == 0:
        warn("%d out of %d %s processed; %2.2f complete"%(ii, iterable_length, object_description, float(ii)/iterable_length))



#!/usr/bin/python

import re
import os
import pdb
import struct
import string
from glob import glob
from random import uniform as runif
import imp, sys
from pathlib import Path

import subprocess
from subprocess import check_call, CalledProcessError

from numpy import array

from math import log, sqrt
from random import random
from scipy.stats import rankdata

from .. import dlaConstants as dlac
from ..mysqlmethods import mysqlMethods as mm

try:
    from PIL import Image
except ImportError:
    dlac.warn("Image module cannot be imported")
    pass

font_ii = 1

def rgb(hex_str):
    """converts a hex string to an rgb tuple"""
    return struct.unpack('BBB', hex_str.decode('hex'))

# def rgbToHex(rgb_tup):
#     #converts an (R, G, B) tuple to Hex code
#     for value in rgb_tup:
#         if value > 255:
#             _warn("An RGB value above 255 found. Capped.")
#             value = 255 #cap value of any RGB value to 255

#     r = "%02X" % rgb_tup[0] #convert decimal to hex
#     g = "%02X" % rgb_tup[1]
#     b = "%02X" % rgb_tup[2]

#     print r
#     print g
#     print b
#     return r + g + b

def findall(string, sep=':'):
    return [m.start() for m in re.finditer(sep, string)]

def explode(string):
    colons = findall(string)
    first = colons[-2]
    second = colons[-1]
    return (string[:first], string[first+1:second], string[second+1:])

def extract(text, sub1, sub2):
    return text.split(sub1)[-1].split(sub2)[0]

def wordcloudByFile(input_filename, output_filename='wordle_test'):
    """Reads a file in format (per line) word:number \n and builds a wordle"""
    word_freq_pairs = []
    f = open(input_filename, 'rb')
    for line in f.readlines():
        word_freq_pairs.append(line.rstrip().split(':'))
    word_list = [x[0] for x in word_freq_pairs]
    freq_list = [x[1] for x in word_freq_pairs]
    wordcloud(word_list, freq_list)

def _makeRandomColorTuples(num_rgb_tuples, rgb_bounds=[0, 0.39]):
    """Create a list of length num_rgb_tuples, where every element is a random hex color"""
    assert(num_rgb_tuples > 0)
    r = list(map(runif, [rgb_bounds[0]]*num_rgb_tuples, [rgb_bounds[1]]*num_rgb_tuples))
    g = list(map(runif, [rgb_bounds[0]]*num_rgb_tuples, [rgb_bounds[1]]*num_rgb_tuples))
    b = list(map(runif, [rgb_bounds[0]]*num_rgb_tuples, [rgb_bounds[1]]*num_rgb_tuples))
    ro_colors = list(map(ro.r.rgb, r, g, b))
    ro_color_list = ro.StrVector([x[0] for x in ro_colors])
    return ro_color_list

def wordcloud(word_list, freq_list, output_prefix='test',
    color_list=None, random_colors=True, random_order=False,
    width=None, height=None, rgb=False, title=None,
    fontFamily="Helvetica-Narrow", keepPdfs=False, fontStyle=None,
    font_path="",
    min_font_size=40, max_font_size=250, max_words=500,
    big_mask=False,
    background_color="#FFFFFF",
    wordcloud_algorithm='ibm'):
    """given a list of words and a list of their frequencies, builds a wordle"""

    PERMA_path = os.path.dirname(os.path.realpath(__file__))
    
    ibm_jar = Path(dlac.DEF_WORDCLOUD_JAR)
    if not ibm_jar.is_file(): wordcloud_algorithm = 'amueller'

    if font_path == "":
        font_path = PERMA_path + "/meloche_bd.ttf"

    if wordcloud_algorithm == 'old': #old wordcloud function
        # requires rpy2, Cairo(rpackage), Wordcloud(rpackage), extrafont(Rpackage), ImageMagick (installed on the system)
        # also: sudo apt-get install libcairo2-dev libxt-dev
        try:
            import rpy2.robjects as ro
            from rpy2.robjects.packages import importr
            ro.r.library('Cairo')       #Optional, only used for old wordcloud module
            ro.r.library('wordcloud')   #Optional, only used for old wordcloud module
            GRDEVICES = importr('grDevices')
            brewer = importr('RColorBrewer')
        except:
            sys.exit('R wordcloud library not imported')


        def _makeColorTuplesFromRgbTuples(rgbTuples):
            r, g, b = list(zip(*rgbTuples))
            N = len(r)
            blank_lists = [ [] for l in range(N) ]
            ro_colors = list(map(ro.r.rgb, r, g, b, [255]*N, blank_lists, [255]*N))
            ro_color_list = ro.StrVector([x[0] for x in ro_colors])
            return ro_color_list

        assert(len(word_list) == len(freq_list))

        if width is None:
            width = 1280
        if height is None:
            height = 800

        size = len(word_list)
        ro_words = ro.StrVector(word_list)
        ro_freqs = ro.FloatVector(freq_list)
        ro_scale = ro.FloatVector([8, 1])

        ro_colors = None
        ordered_colors = False
        if color_list:
            if rgb:
                ro_colors = _makeColorTuplesFromRgbTuples(color_list)
            else:
                ro_colors = ro.StrVector(color_list)
                if not random_colors: ordered_colors = True

        else:
            ro_colors = _makeRandomColorTuples(size)

        if output_prefix:
            # (if Cairo gives an error you need to install it locally through R: install.packages("Cairo"))
            if fontStyle:
                fontFamily = fontFamily + ':' + fontStyle
            ro.r.CairoFonts(regular=fontFamily)
            ro.r.Cairo(type="pdf", file="%s_wc.pdf"%output_prefix, width=11, height=8.5, units="in", onefile=True)
        ro.r.options(warn=2)
        while True:
            try:
                ro.r.wordcloud(ro_words, ro_freqs, rot_per=0, scale=ro_scale,max_words=25,random_order=random_order,colors=ro_colors, random_color=random_colors)
                if title:
                    ro.r.text(0, 1, labels=title, cex=0.8, pos=4)
                break
            except Exception as e:
                ro_scale[0] -= 1
                if ro_scale[0] <= ro_scale[1]:
                    raise Exception("Fatal Error: Either Words do not fit in wordle or error is due to R: [%s]"%e.message)
        ro.r.options(warn=0)
        if output_prefix:
            GRDEVICES.dev_off()
            # count number of pdf pages
            pdfFile = output_prefix + "_wc.pdf"
            nPages = -1
            with open(pdfFile, 'rb') as rf:
                for line in rf.readlines():
                    if "/Count " in line:
                        nPages = int( re.search("/Count \d*", line).group()[7:] )
                        # print pdfFile, 'has', nPages, 'pages'
            # convert the last page to png
            pngFile = output_prefix + "_wc.png"
            try:
                check_call(["convert", "-quality", "100", pdfFile + "[%d]"%(nPages-1), pngFile])
            except (OSError, CalledProcessError) as e:
                print('ERROR:', e)
            if not keepPdfs:
                check_call(["rm", pdfFile])

    elif wordcloud_algorithm == 'amueller':   #new wordcloud function
        f, pathname, desc = imp.find_module('wordcloud', sys.path[1:])
        wc = imp.load_module('wc', f, pathname, desc)
        if f is not None:
            f.close()
        #explicitly import amueller's wordcloud library
        #for all intents that block above is equivalent to: import wordcloud as wc
        #this is needed as both the local module and amueller's package are both called 'wordcloud'

        assert(len(word_list) == len(freq_list))
        color_string_list = None
        if width is None:
            width = 1280
        if height is None:
            height = 800

        print('Generating wordclouds using amueller\'s python package...')

        word_freq_tup = list(zip(word_list, freq_list))
        #list of tups of (word, freq)

        word_freq_tup = sorted(word_freq_tup, key=lambda tup: -tup[1]) #sort by frequency


        if color_list:
            #color list is list of colors to be applied to words with least to most freq
            assert(len(color_list) == len(word_list))
            if rgb:
                color_string_list = ["rgb(%d,%d,%d)" % (x[0], x[1], x[2]) for x in color_list]
                #print color_string_list
            else:
                color_string_list = color_list #assume that the color list already has usable color strings
                if not random_colors: ordered_colors = True

        else:
            color_string_list = _makeRandomColorTuples(len(word_list))
            #create list of random color tuples, one for each word

        sorted_word_list, sorted_freq_list = list(zip(*word_freq_tup))



        color_dict = {sorted_word_list[i]: color_string_list[i] for i in range(0, len(sorted_word_list))}
        #create dictionary where each word is mapped to a rgb value

        def color_func(word=None, font_size=None, position=None,
                      orientation=None, font_path=None, random_state=None):
            try:
                return color_dict[word]
            except ValueError:
                return 'black'

        img_dir = os.path.dirname(os.path.abspath(__file__))
        if (big_mask is True or len(word_list) > 75): #arbitrary number for choosing mask
            img = Image.open(os.path.join(img_dir, 'oval_big_mask.png'))
        else:
            img = Image.open(os.path.join(img_dir, 'oval_mask.png'))

        img_array = array(img) #nd-array

        cloud = wc.WordCloud(font_path=font_path,
                    width=width,
                    height=height,
                    max_font_size=max_font_size,
                    min_font_size=min_font_size,
                    max_words=max_words,
                    color_func=color_func,
                    prefer_horizontal=1.0,
                    background_color=background_color,
                    mask=img_array).generate_from_frequencies(word_freq_tup)

        if output_prefix:
            #TODO: pdf output?
            pngFile = output_prefix + "_wc.png"
            cloud.to_file(pngFile)
            print(('Wordcloud created at: %s' % pngFile))
        else:
            warn('No filename specified. Wordcloud not created.')
            #Hey buddy. You didn't specify the filename.

    elif wordcloud_algorithm == 'ibm':
        if width is None:
            width = 1000
        if height is None:
            height = 2000

        if background_color[0] == "#": #TODO: do an actual HEX check
            background_color = background_color[1:]

        config_loc = PERMA_path + "/wc_config.txt";
        #if not os.path.isfile(config_loc):
        with open(config_loc, "w") as f:
            f.write('font: ' + font_path + '\n')
            f.write("format: tab\n" +
                    "inputencoding: UTF-8\n" +
                    "firstline: data\n" +
                    "wordcolumn: 1\n" +
                    "weightcolumn: 2\n" +
                    "colorcolumn: 3\n" +
                    "background: " + background_color + "\n" +
                    "placement: HorizontalCenterLine\n" +
                    "shape: BLOBBY\n" +
                    "orientation: HORIZONTAL")

        if not output_prefix:
            warn('No filename specified. Filename specified as \'wc\'.')
            output_prefix = 'wc'

        words_loc = os.path.normcase(output_prefix)
        words_loc = words_loc + '.txt'

        # if (output_prefix[:2] == './' or output_prefix[:1] == '/' ):
        #     words_loc = output_prefix
        # else: #assuming relative path
        #     words_loc = "./" + output_prefix

        # if (output_prefix[-1:] == '/'):
        #     words_loc = words_loc[:-1] + '.txt'
        # else:
        #     words_loc = words_loc + '.txt'


        with open(words_loc, 'w') as f:
            word_freq_tup = list(zip(word_list, freq_list))
            word_freq_tup = sorted(word_freq_tup, key=lambda tup: -tup[1]) #sort by frequency

            if color_list:
                #color list is list of colors to be applied to words with least to most freq
                assert(len(color_list) == len(word_list))
                if rgb:
                    color_string_list = ["rgb(%d,%d,%d)" % (x[0], x[1], x[2]) for x in color_list]
                    #print color_string_list
                else:
                    color_string_list = color_list #assume that the color list already has usable color strings
                    if not random_colors: ordered_colors = True

            else:
                color_string_list = _makeRandomColorTuples(len(word_list))
                #create list of random color tuples, one for each word

            sorted_word_list, sorted_freq_list = list(zip(*word_freq_tup))
            word_freq_color_tup = list(zip(sorted_word_list, sorted_freq_list, color_string_list))


            pngFile = output_prefix + '.png'

            for tup in word_freq_color_tup:
                try:
                    word_row = tup[0] + '\t' + str(tup[1]) + '\t' + str(tup[2]) + '\n'
                    f.write(word_row)
                except Exception as e:
                    print(e)
                    print("Line contains unprintable unicode, skipped: ", end=' ')
                    print(tup)
                #str is necessary in case the words are numbers

        command = ['java','-jar', str(ibm_jar), '-c',
                    config_loc, '-w', str(width), '-h', str(height), '-i',
                    words_loc, '-o', pngFile]
        print('Generating wordcloud at ' + pngFile)

        #print ' '.join(command)
        with open(os.devnull, 'w') as fnull: #for suppressing outputs of the command
            subprocess.call(command, stdout=fnull, stderr=subprocess.STDOUT)

    else:
        print("Wordcloud algorithm not recognized.")

def _tagcloudToWordcloud(filename='', directory=os.getcwd()):
    f = open(os.path.join(directory, filename), 'rb')
    lastline = ''
    last_starline = False
    bracketline = False
    getwords = False
    tagTriples = []
    for line in f.readlines():
        line = line.strip()
        if getwords:
            if not line:
                getwords = False
                topicId = extract(lastline, 'Topic Id:',',').strip()
                # tags = map(lambda x:(x[0].upper(), int(x[1]), rgb(x[2])), tagTriples)
                tags = [(x[0].lower(), int(x[1]), rgb(x[2])) for x in tagTriples]
                ngrams, freqs, colors = list(zip(*tags))
                lastline = ''
                tagTriples = []
#                print topicId, ngrams, freqs, colors
                wordcloud(ngrams, freqs, os.path.join(directory, topicId),colors,rgb=True)
                break
            else:
                tagTriples.append(explode(line))

        if len(line) > 0 and line[0] == '[' and line[-1] == ']':
            bracketline = True
            lastline = line
            if not last_starline:
                getwords = True
            bracketline = False
        if len(line) > 1 and line[0:2] == '**' and line[-2:]:
            last_starline = True
        else:
            last_starline = False

    f.close()

def processTopicLine(line, metric):
    splitLine = line.split(' ')
    return 'tid-'+splitLine[2][0:-1], metric + '-'+splitLine[4][0:-1][0:5], '' #'tFq'+splitLine[6][0:-1]

def processWordCloudLine(line, metric):
    splitLine = line.split(' ')
    maxr_idx, minr_idx = 2, 5
    maxR, minR = splitLine[maxr_idx][0:-1], splitLine[minr_idx][0:-1]
    return maxR, minR

def duplicateFilterLineIntoInformativeString(line):
    splitLine = line.split(' ')
    return 'df-%s-of-%s'%(splitLine[1], splitLine[5])

def coerceToValidFileName(filename):
    valid_chars = "-_.() %s%s" % (string.ascii_letters, string.digits)
    return ''.join(c for c in filename if c in valid_chars)

def tagcloudToWordcloud(filename='', directory='', withTitle=False, fontFamily="Helvetica-Narrow", fontStyle=None, toFolders=False, useTopicWords=True, metric='R', keepDuplicates=False):
    processedFiles = {}

    if not directory:
        directory = os.path.dirname(filename)
        filename = os.path.basename(filename)

    fileFolders = {}
    filename = filename + '.txt'
    basefilename = filename[:-4]
    # basefolder = os.path.join(directory, basefilename + '.output-dir')
    basefolder = os.path.join(directory, basefilename + '_wordclouds')

    with open(os.path.join(directory, filename), 'rb') as f:
        lastline = ''; tagTriples = [];
        getwords = False; posneg=''; isNewTagcloud = False;
        outcome = ''; topicId = None; topicR = None; topicFreq = None;
        topicDup = '';
        topicFiles = False
        maxR, minR = None, None

        for line in f.readlines():
            line = line.strip().decode("utf-8")

            if getwords:
                if not line and len(tagTriples)==0:
                    getwords = False
                elif line and len(line) >= 19 and line[0:19]=='rList has less than':
                    getwords = False
                elif line and len(line) >= 18 and line[0:18]=='Error_in_function_':
                    line = None
                elif line and len(line) >= 28 and line[0:28]=='rList has less than no items':
                    getwords = False
                elif '(converted_from_warning)' in line:
                    line = None
                elif line:
                    tagTriples.append(explode(line))

                else:
                    tags = [(x[0].lower(), int(x[1]), x[2]) for x in tagTriples]
                    newFilename = None
                    fileEnding = ''
                    fileEnding += '.' + outcome
                    fileEnding += '.' + posneg
                    if topicR: fileEnding += '.' + topicR
                    if topicId: fileEnding += '.' + topicId
                    if useTopicWords:
                        ngrams, freqs, colors = list(zip(*tags))
                        fileEnding += '.'.join(ngrams[0:6])
                    if topicFreq: fileEnding += '.' + topicFreq
                    if topicDup: fileEnding += '.' + topicDup
                    fileEnding += '.txt'
                    newFilename = filename.replace('.txt', fileEnding) if '.txt' in filename else filename + fileEnding
                    newFilename = coerceToValidFileName(newFilename)
                    processedFiles[newFilename[0:-4]] = tags
                    topicR = '' if not topicR else topicR
                    topicId = '' if not topicId else topicId
                    topicDup = '' if not topicDup else topicDup
                    topicFreq = '' if not topicFreq else topicFreq

                    if toFolders:
                        ngrams, freqs, colors = list(zip(*tags))
                        ngramString = '.'.join(ngrams[0:min(6, len(ngrams))])
                        ngramString = '' ## LAD delete this line to add ngrams / entries back into the file
                        fileFolders[newFilename[0:-4]] = (outcome, '.'.join(filter(None, [posneg, topicR, topicId, ngramString, topicDup, topicFreq])))

                    getwords = False; tagTriples = [];
                    topicDup = ''; topicId = None; topicR = None; topicFreq = None;


            if line and len(line) >= 10 and line[0:10] == '[Topic Id:':
                topicId, topicR, topicFreq = processTopicLine(line, metric)
                getwords = True
                topicFiles = True
            if line and len(line)>=6 and line[0:3]=='**[' and line[-3:]==']**':
                topicDup = duplicateFilterLineIntoInformativeString(line)
            if line and len(line)==12 and line == '------------':
                posneg = 'pos'; 
            if line and len(line)==13 and line == '-------------':
                posneg = 'neg';
            if line and line.startswith('[Max %s: ' % metric):
                maxR, minR = processWordCloudLine(line, metric)
                topicR = metric + '_' + minR + '-' + maxR
                getwords = True;
            if line and len(line)==27 and line[0:27] == '---------------------------':
                outcome = lastline[19:];
            if line and len(line)==32  and line[0:32] == '--------------------------------':
                outcome = lastline[25:]
            lastline = line

        if getwords:
            tags = [(x[0], int(x[1]), x[2]) for x in tagTriples]
            if tags:
                # ngrams, freqs, colors = zip(*tags)
                newFilename = None
                fileEnding = ''
                fileEnding += '.' + outcome
                fileEnding += '.' + posneg
                if topicR: fileEnding += '.' + topicR
                if topicId: fileEnding += '.' + topicId
                if useTopicWords:
                    ngrams, freqs, colors = list(zip(*tags))
                    fileEnding += '.'.join(ngrams[0:6])
                if topicFreq: fileEnding += '.' + topicFreq
                if topicDup: fileEnding += '.' + topicDup
                fileEnding += '.txt'
                newFilename = filename.replace('.txt', fileEnding) if '.txt' in filename else filename + fileEnding
                processedFiles[newFilename[0:-4]] = tags
                topicR = '' if not topicR else topicR
                topicId = '' if not topicId else topicId
                topicDup = '' if not topicDup else topicDup
                topicFreq = '' if not topicFreq else topicFreq


                newFilename = coerceToValidFileName(newFilename)
                if toFolders:
                    ngrams, freqs, colors = list(zip(*tags))
                    ngramString = '.'.join(ngrams[0:min(6, len(ngrams))])
                    ngramString = '' ## LAD delete this line to add ngrams / entries back into the file
                    suffixList = filter(None, [posneg, topicR, topicId, ngramString, topicDup, topicFreq])
                    fileFolders[newFilename[0:-4]] = (outcome, '.'.join(suffixList))

    # from pprint import pprint as pp
    ii = 0

    # import pdb
    # pdb.set_trace()

    if toFolders:
        try:
            os.mkdir(basefolder)
        except OSError:
            pass

    for filename, tags in processedFiles.items():
        if not keepDuplicates and '.df-' in filename: continue
        ngrams, freqs, colors = list(zip(*tags))
        colors = ['#'+x for x in colors]
        filename = coerceToValidFileName(filename)
        title = filename if withTitle else None
        output_file = None
        if toFolders:
            outcome, endname = fileFolders[filename]
            if topicFiles:
                try:
                    os.mkdir(os.path.join(basefolder, outcome))
                except OSError:
                    pass
                output_file = os.path.join(basefolder, outcome, endname)
            else:
                try:
                    os.mkdir(os.path.join(basefolder))
                except OSError:
                    pass
                output_file = os.path.join(basefolder, outcome+'_'+endname)

        else:
            output_filez = os.path.join(directory, filename)
        wordcloud(ngrams, freqs,output_file, colors, rgb=False, title=title, fontFamily=fontFamily, fontStyle=fontStyle)
        try:
            wordcloud(ngrams, freqs,output_file, colors, rgb=False, title=title, fontFamily=fontFamily, fontStyle=fontStyle)
        except Exception as e:
            print('WARNING: ERROR happened for file: %s'%(output_file,))
            print(e)

        # ii += 1
        # if ii > 2:
        #     break

    return processedFiles

# wordcloud tools
def freqToColor(freq, maxFreq = 1000, resolution=64, colorScheme='multi'):
    """
    Alter color scheme of plot based on the the word frequencies

    Parameters
    ----------
    freq : int
        word frequency
    maxFreq : :obj:`int`, optional
        maximum frequency threshold
    resolution : :obj:`int`, optional
        pixels of resolution
    colorScheme : :obj:`str`, optional
        specifies color scheme of plot

    Returns
    -------
    htmlcode : str
        string of html code specifying color scheme


    """
    perc = freq / float(maxFreq)
    (red, green, blue) = (0, 0, 0)
    if colorScheme=='multi':
        if perc < 0.17: #grey to darker grey
            (red, green, blue) = dlac.rgbColorMix((168, 168, 168),(124, 124, 148), resolution)[int(((1.00-(1-perc))/0.17)*resolution) - 1]
        elif perc >= 0.17 and perc < 0.52: #grey to blue
            (red, green, blue) = dlac.rgbColorMix((124, 124, 148), (32, 32, 210), resolution)[int(((0.830-(1-perc))/0.35)*resolution) - 1]
        elif perc >= 0.52 and perc < 0.90: #blue to red
            (red, green, blue) = dlac.rgbColorMix((32, 32, 210), (200, 16, 32), resolution)[int(((0.48-(1-perc))/0.38)*resolution) - 1]
        else: #red to dark red
            (red, green, blue) = dlac.rgbColorMix((200, 16, 32), (128, 0, 0), resolution)[int(((0.10-(1-perc))/0.10)*resolution) - 1]
    # blue:
    elif colorScheme=='blue':
        if perc < 0.50: #light blue to med. blue
            (red, green, blue) = dlac.rgbColorMix((76, 76, 236), (48, 48, 156), resolution)[int(((1.00-(1-perc))/0.5)*resolution) - 1]
        else: #med. blue to strong blue
            (red, green, blue) = dlac.rgbColorMix((48, 48, 156), (0, 0, 110), resolution)[int(((0.5-(1-perc))/0.5)*resolution) - 1]
    elif colorScheme=='old_blue':
        if perc < 0.50: #light blue to med. blue
            (red, green, blue) = dlac.rgbColorMix((76, 76, 236), (48, 48, 156), resolution)[int(((1.00-(1-perc))/0.5)*resolution) - 1]
        else: #med. blue to strong blue
            (red, green, blue) = dlac.rgbColorMix((48, 48, 156), (0, 0, 110), resolution)[int(((0.5-(1-perc))/0.5)*resolution) - 1]
    #red:
    elif colorScheme=='red':
        if perc < 0.50: #light red to med. red
            (red, green, blue) = dlac.rgbColorMix((236, 76, 76), (156, 48, 48), resolution)[int(((1.00-(1-perc))/0.5)*resolution) - 1]
        else: #med. red to strong red
            (red, green, blue) = dlac.rgbColorMix((156, 48, 48), (110, 0, 0), resolution)[int(((0.5-(1-perc))/0.5)*resolution) - 1]
    elif colorScheme=='green':
        (red, green, blue) = dlac.rgbColorMix((166, 247, 178), (27, 122, 26), resolution)[int((1.00-(1-perc))*resolution) - 1]
    elif colorScheme=='teal':
        (red, green, blue) = dlac.rgbColorMix((198, 255, 255), (0, 128, 128), resolution)[int(((0.5-(1-perc))/0.5)*resolution) - 1]
    #red+randomness:
    elif colorScheme=='red-random':
        if perc < 0.50: #light blue to med. blue
            (red, green, blue) = dlac.rgbColorMix((236, 76, 76), (156, 48, 48), resolution, True)[int(((1.00-(1-perc))/0.5)*resolution) - 1]
        else: #med. blue to strong blue
            (red, green, blue) = dlac.rgbColorMix((156, 48, 48), (110, 0, 0), resolution, True)[int(((0.5-(1-perc))/0.5)*resolution) - 1]
    else:
        dlac.warn('Warning: "%s" colorscheme not found, defaulting to black' % (colorScheme))
    htmlcode = "%02s%02s%02s" % (hex(red)[2:], hex(green)[2:], hex(blue)[2:])
    return htmlcode.replace(' ', '0')

def getRankedFreqList(word_list, max_size = 75, min_size = 30, scale = 'linear'):
    """
    returns freq_list i.e. list of sizes from word_list
    freq_list goes from biggest to smallest
    make sure the word_list is sorted accordingly
    """
    if len(word_list) == 1:
        return [max_size]

    freq_list = []

    num_blocks = int(log(len(word_list), 2) + 1)
    range = max_size - min_size
    block_size = range/(num_blocks - 1)


    i = 1;
    while i <= len(word_list):
        rank = int(log(i, 2))
        #print "{} {}".format(i, rank)

        if scale == 'linear':
            value = max_size - (block_size * rank)
            freq_list.append(value)


        i += 1

    return freq_list

def normalizeFreqList(old_freq_list, word_count = 15):
    """
    Given a sorted freq_list and a word count, return a normalized freq_list, based on the old sizing algorithm from oa.printTagCloudFromTuples
    :param old_freq_list: list of sorted, descending integers
    :param word_count: an integer that shows how big the new_freq_list should be
    """

    minR = old_freq_list[-1]
    maxR = old_freq_list[0]
    diff = float(maxR - minR)
    if diff == 0: diff = 0.000001
    smallDataBump = max((word_count - len(old_freq_list)), 10)

    new_freq_list = [int(((freq-minR)/diff)*word_count) + smallDataBump for freq in old_freq_list]

    return new_freq_list

def getColorList(word_list, freq_list = [], randomize = False, colorScheme = 'multi', scale = 'linear'):
    color_list = []

    max_freq = 1000
    for i in range(len(word_list)):
        if randomize:
            #print 'Randomizing colors'
            freq = (random() * max_freq) + 1 #a number from 1 to max_freq
            colorHex = freqToColor(freq, maxFreq = max_freq, colorScheme = colorScheme)
            color_list.append(colorHex)

        elif freq_list:
            assert (len(word_list) == len(freq_list))
            rank_list = rankdata(freq_list, method = 'ordinal') #, method = 'ordinal'
            #print rank_list
            if scale == 'sqrt':
                max_size = max([sqrt(x) for x in freq_list]) #scaled via sqrt
                freq = sqrt(freq_list[i])

            elif scale == 'linear':
                max_size = len(freq_list)
                freq = rank_list[i]

            #max_size = min([max_size, 100])
            #freq = min(freq_list[i], 100)
            colorHex = freqToColor(freq, maxFreq = max_size, colorScheme = colorScheme)
            color_list.append(colorHex)
            #print '{} {}'.format(word_list[i], max_size)
        else:
            print('Randomize is False, but freq_list is not provided.')



    return color_list

def getFeatValueAndZ(user, schema, ngramTable, min_value = 5, ordered = True, z_threshold = 0):
    """returns list of (feat, value, z) for a given user"""
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)

    if ordered:
        order_by = " ORDER BY z DESC"
    else:
        order_by = ""

    pos_z = " AND z > {}".format(z_threshold)


    query = 'SELECT feat, value, z FROM {}.{} WHERE group_id = \'{}\' and value >= {}{}{};'.format(schema, ngramTable, user, min_value, pos_z, order_by)
    print(query)
    list = mm.executeGetList(schema, dbCursor, query)
    return list


def getMeanAndStd(word, ngramTable, schema, num_groups = -1, distTable = '', distTableSource = None):
    """get mean and std for a word using the ngramTable"""

    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)

    if num_groups == -1:
        query = 'SELECT count(distinct(group_id)) FROM {}.{}'.format(schema, ngramTable)
        result = mm.executeGetList(schema, dbCursor, query)
        num_groups = int(result[0][0])
        #print int(num_groups[0][0])

    elif distTableSource is not None:
        #TODO: let user specify distTableSource
        query = 'SELECT count(distinct(group_id)) FROM {}.{}'.format(schema, distTableSource)
        result = mm.executeGetList(schema, dbCursor, query)
        num_groups = int(result[0][0])
    else:
        pass

    if distTable == '':

        ########### two pass algorithm
        n = 0 #count
        sum = 0.0
        diff_squared_sum = 0.0
        #for group_norm in session.query(Feature.group_norm).filter(Feature.feat == word):
        query = 'SELECT group_norm FROM {}.{} WHERE feat = \'{}\''.format(schema, ngramTable, word)
        group_norms = mm.executeGetList(schema, dbCursor, query)
        #print 'SELECT group_norm FROM {}.{} WHERE feat = \'{}\''.format(schema, ngramTable, word)
        num_groups = len(group_norms)

        if len(group_norms) == 1:
            return (0, 0)

        for group_norm in group_norms:
            n += 1
            sum += group_norm[0]


        mean = float(sum)/num_groups
        #print "Mean: %.12f" % mean

        for group_norm in group_norms:
            diff_squared_sum += (group_norm[0] - mean) ** 2

        if (num_groups == 1):
            variance = 1
        else:
            variance = diff_squared_sum / (num_groups - 1) #sample variance
        std = sqrt(variance)
        #print "Standard Deviation: %.12f" % std

        ########### algorithm end

    else:
        query = "SELECT mean, std FROM {}.{} where feat = \'{}\'".format(schema, distTable, word)
        result = mm.executeGetList(schema, dbCursor, query)
        if not result:
            mean = 0
            std = 0
        else:
            mean = result[0][0]
            std = result[0][1]

    return (mean, std)

def getNgrams(ngramTable, schema):
    #returns list of ngrams

    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)
    query = "SELECT feat FROM {}.{} GROUP BY feat".format(schema, ngramTable)
    return mm.executeGetList(schema, dbCursor, query)

def getUsers(schema, ngramTable):
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)
    query = "SELECT distinct(group_id) FROM {}.{};".format(schema, ngramTable)
    return [user[0] for user in mm.executeGetList(schema, dbCursor, query)]

def updateZscore(schema, ngramTable, user = '', use_feat_table = False, distTable = ''):
    # update ngramTable with z-values

    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)

    counter = 0
    if user != '':
        users = [user]

    else:
        users = getUsers(schema, ngramTable)

    for user in users:
        for ngram in [x[0] for x in getNgrams(ngramTable, schema)]:
            if use_feat_table:
                z = getZscore(ngram, user, ngramTable, schema)
            else:
                z = getZscore(ngram, user, ngramTable, schema, distTable = distTable)

            ngram = ngram.replace('\'', '\'\'')

            try :
                query = "UPDATE {}.{} SET z = {} where group_id = \'{}\' and feat=\'{}\'".format(schema, ngramTable, z, user, ngram)

            except UnicodeEncodeError:
                query = "UPDATE {}.{} SET z = 0 where group_id = \'{}\' and feat=\'{}\'".format(schema, ngramTable, user, ngram.encode('utf-8'))


            if counter % 1000 == 0: print(query)
            mm.executeGetList(schema, dbCursor, query)
            counter += 1

def getZscore(word, user, ngramTable, schema, distTable = ''):
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)
    word = word.replace('\'', '\'\'')

    try:
        query = 'SELECT group_norm FROM {}.{} where group_id = \'{}\' and feat = \'{}\''.format(schema, ngramTable, user, word)
    except UnicodeEncodeError:
        return 0

    group_norm = mm.executeGetList(schema, dbCursor, query)

    if not group_norm:
        return 0
    if isinstance(group_norm, tuple):
        #print group_norm
        group_norm = group_norm[0]

    if isinstance(group_norm, tuple):
        group_norm = group_norm[0]

    (mean, std) = getMeanAndStd(word, ngramTable = ngramTable, schema = schema, distTable = distTable)
    if (std == 0):
        return 0
    else:
        return (group_norm - mean)/(std + 0.0)

def createZColumn(schema, ngramTable):
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)
    query = "ALTER TABLE {}.{} ADD COLUMN z DOUBLE;".format(schema, ngramTable)
    mm.executeGetList(schema, dbCursor, query)

def getOneGram(schema, ngramTable):
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)
    query = "SELECT feat, sum(value) as count FROM {}.{} group by feat".format(schema, ngramTable)
    print(query)
    return mm.executeGetList(schema, dbCursor, query)

def getUniqueNgrams(schema, ngramTable, user = '', max = -1):
    """
    get n ngrams from ngramTable where z-score = 0, sorted by group_norm
    if user is specified, only grab unique ngrams from that user
    """
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)

    if user != '':
        select_user = ' AND group_id = \'{}\''.format(user)
    else:
        select_user = ''

    if max != -1:
        limit = ' LIMIT {}'.format(max)
    else:
        limit = ''

    query = 'SELECT feat, group_norm FROM {}.{} WHERE z = 0{} ORDER BY group_norm DESC{}'.format(schema, ngramTable, select_user, limit)
    return mm.executeGetList(schema, dbCursor, query)

def getFeatWithLimit(schema, table, group = '', amount = 50, orderBy = 'group_norm', desc = True):
    """
    get the first n amount of words, using the orderBy (asc or desc) column to sort.
    if group is specified, get from that specific group
    returns list of (feat, group_norm)
    """
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(schema)

    if group != '':
        select_group = 'where group_id = \'{}\''.format(group)
    else:
        select_group = ''

    if amount <= 0:
        limit = ''
    else:
        limit = ' LIMIT {}'.format(int(amount))

    query = 'SELECT feat, group_norm FROM {}.{} {} ORDER BY {} DESC{}'.format(schema, table, select_group, orderBy, limit)
    return mm.executeGetList(schema, dbCursor, query)

# wordcloud print methods
def makeLexiconTopicWordclouds(lexdb, lextable, output, color, max_words=15, cleanCloud=False, censor_dict=dlac.DEF_CENSOR_DICT):
    if not os.path.exists(output):
        os.makedirs(output)
    (dbConn, dbCursor, dictCursor) = mm.dbConnect(lexdb)
    query = """SELECT distinct(category) FROM %s.%s""" % (lexdb, lextable)
    categories = map(lambda x: x[0], mm.executeGetList(lexdb, dbCursor, query))


    for category in categories:
        query = """SELECT term, weight FROM %s.%s WHERE category = \'%s\' ORDER BY weight desc LIMIT %d""" % (lexdb, lextable, category, max_words)
        result = mm.executeGetList(lexdb, dbCursor, query)
        (word_list, freq_list) = zip(*result)

        ranked_freq_list = normalizeFreqList(freq_list, word_count = max_words)
        color_list = getColorList(word_list, freq_list = ranked_freq_list, colorScheme = color)

        if cleanCloud:
            word_list = tuple([censor_dict[t] if t in censor_dict.keys() else t for t in word_list])


        output_name = os.path.join(output, 'topic_' + str(category))
        wordcloud(word_list, ranked_freq_list, color_list = color_list, output_prefix = output_name)



if __name__=='__main__':
    print('wordcloud sleeps peacefully in the shade, its work is happily done.')



import sys
import time
import MySQLdb

from . import dlaConstants as dlac
from .mysqlmethods import mysqlMethods as mm 

class DLAWorker(object):
    """Generic class for functions working with features

    Parameters
    ----------
    corpdb : str
        Corpus Database Name.
    corptable : str
        Corpus Table.
    correl_field : str
        Correlation Field (AKA Group Field): The field which features are aggregated over.
    mysql_host : str
        Host that the mysql server runs on.
    message_field : str
        The field where the text to be analyzed is located.
    messageid_field : str
        The unique identifier for the message.
    encoding : str
        MySQL encoding
    lexicondb : :obj:`str`, optional
        The database which stores all lexicons.
    date_field : :obj:`str`, optional
        Date a message was sent (if avail, for timex processing).
    wordTable : :obj:`str`, optional
        Table that contains the list of words to give for lex extraction/group_freq_thresh

    Returns
    -------
    DLAWorker object
    """
    
    def __init__(self, corpdb, corptable, correl_field, mysql_host, message_field, messageid_field, encoding, use_unicode, lexicondb = dlac.DEF_LEXICON_DB, date_field=dlac.DEF_DATE_FIELD, wordTable=None):
        self.corpdb = corpdb
        self.corptable = corptable
        self.correl_field = correl_field
        self.mysql_host = mysql_host
        self.message_field = message_field
        self.messageid_field = messageid_field
        self.encoding = encoding
        self.use_unicode = use_unicode
        (self.dbConn, self.dbCursor, self.dictCursor) = mm.dbConnect(corpdb, host=mysql_host, charset=encoding)
        self.lexicondb = lexicondb
        self.wordTable = wordTable if wordTable else "feat$1gram$%s$%s$16to16"%(self.corptable, self.correl_field)
        self.messageIdUniqueChecked = False

    ##PUBLIC METHODS#
    def checkIndices(self, table, primary=False, correlField=False):
        hasPrimary, hasCorrelIndex = True, True
        warn_message = "WARNING: The table %s does not have:"  % table
        if primary:
            hasPrimary = mm.primaryKeyExists(self.dbConn, self.dbCursor, table, correlField)
            if not hasPrimary: warn_message += " a PRIMARY key on %s" % correlField
        elif correlField:
            hasCorrelIndex = mm.indexExists(self.dbConn, self.dbCursor, table, correlField)
            if not hasCorrelIndex: 
                if not hasPrimary: warn_message += " or"
                warn_message += " an index on %s" % correlField
        warn_message += ". Consider adding."
        if self.messageid_field == correlField and not hasPrimary:
            warn_message += "\n         Please check that all messages have a unique %s, this can significantly impact all downstream analysis" % (self.messageid_field)
        if not hasPrimary or not hasCorrelIndex:
            dlac.warn(warn_message)

    def getMessages(self, messageTable = None, where = None):
        """?????
 
        Parameters
        ----------
        messageTable : :obj:`str`, optional
            Name of message table.
        where : :obj:`str`, optional
            Filter groups with sql-style call.
     
        Returns
        -------
        ?????
            ?????
        """
        if not messageTable: messageTable = self.corptable
        msql = """SELECT %s, %s FROM %s"""% (self.messageid_field, self.message_field, messageTable)
        if where: msql += " WHERE " + where
        return mm.executeGetSSCursor(self.corpdb, msql, charset=self.encoding, host=self.mysql_host)

    def getMessagesForCorrelField(self, cf_id, messageTable = None, warnMsg = True):
        """?????
 
        Parameters
        ----------
        cf_id : str
            Correl field id.
        messageTable : :obj:`str`, optional
            name of message table.
        warnMsg : :obj:`boolean`, optional
            ?????
     
        Returns
        -------
        ?????
            ?????
        """
        if not messageTable: messageTable = self.corptable
        # check that self.messageid_field is unique
        if self.messageIdUniqueChecked == False:
            self.checkIndices(messageTable, primary=True, correlField=self.messageid_field)
            if self.correl_field != self.messageid_field:
                self.checkIndices(messageTable, primary=False, correlField=self.correl_field)
            self.messageIdUniqueChecked = True
        msql = """SELECT %s, %s FROM %s WHERE %s = '%s'""" % (
            self.messageid_field, self.message_field, messageTable, self.correl_field, cf_id)
        #return self._executeGetSSCursor(msql, warnMsg, host=self.mysql_host)
        return mm.executeGetList(self.corpdb, self.dbCursor, msql, warnMsg, charset=self.encoding)

    def getMessagesWithFieldForCorrelField(self, cf_id, extraField, messageTable = None, warnMsg = True):
        """?????
 
        Parameters
        ----------
        cf_id : str
            Correl field id.
        extraField : str
            ?????
        messageTable : :obj:`str`, optional
            name of message table.
        warnMsg : :obj:`boolean`, optional
            ?????
     
        Returns
        -------
        describe : list
            A list of messages for a given correl field id.
        """
        if not messageTable: messageTable = self.corptable
        msql = """SELECT %s, %s, %s FROM %s WHERE %s = '%s'""" % (
            self.messageid_field, self.message_field, extraField, messageTable, self.correl_field, cf_id)
        #return self._executeGetSSCursor(msql, showQuery)
        return mm.executeGetList(self.corpdb, self.dbCursor, msql, warnMsg, charset=self.encoding)

    def getMidAndExtraForCorrelField(self, cf_id, extraField, messageTable = None, where = None, warnMsg = True):
        """?????
 
        Parameters
        ----------
        cf_id : str
            Correl field id.
        extraField : str
            ?????
        messageTable : :obj:`str`, optional
            name of message table.
        warnMsg : :obj:`boolean`, optional
            ?????
     
        Returns
        -------
        describe : list
            A list of messages for a given correl field id.
        """
        if not messageTable: messageTable = self.corptable
        msql = """SELECT %s, %s FROM %s WHERE %s = '%s'""" % (
            self.messageid_field, extraField, messageTable, self.correl_field, cf_id)
        if where:
            msql += ' AND '+where
        #return self._executeGetSSCursor(msql, showQuery)
        return mm.executeGetList(self.corpdb, self.dbCursor, msql, warnMsg, charset=self.encoding)

    
    def getNumWordsByCorrelField(self, where = ''):
        """?????
 
        Parameters
        ----------
        where : :obj:`str`, optional
            Filter groups with sql-style call.
     
        Returns
        -------
        ?????
            ?????
        """
        #assumes corptable has num_words field for each message
        #SELECT user_id, sum(num_words) FROM (SELECT user_id, num_words FROM messages GROUP BY message_id) as a GROUP BY user_id
        sql = """SELECT %s, sum(num_words) FROM (SELECT %s, num_words FROM %s """ % (self.correl_field, self.correl_field, self.corptable)
        if (where): sql += ' WHERE ' + where  
        sql += """ GROUP BY %s) as a """ % self.messageid_field 
        sql += """ GROUP BY %s """ % self.correl_field
        return mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

    def getWordTable(self, corptable = None):
        """?????
 
        Parameters
        ----------
        corptable : :obj:`str`, optional
            Choices in brackets, default first when optional.
     
        Returns
        -------
        str
            Name of word table for given corptable and correl_field.
        """
        if self.wordTable: return self.wordTable
        if not corptable:
            corptable  = self.corptable
        return "feat$1gram$%s$%s$16to16"%(corptable, self.correl_field)

    def get1gramTable(self):
        """?????
     
        Returns
        -------
        str
            Name of word table (1gram table) for a given corptable and correl field.
        """
        return "feat$1gram$%s$%s$16to16"%(self.corptable, self.correl_field)
        
    def getWordTablePOcc(self, pocc):
        """This function does something.
 
        Parameters
        ----------
        pocc : float
            Value of p occurrence filter.
     
        Returns
        -------
        str
            A word table (1gram table) for a given corptable, correl field and p occurrence value.
        """
        return "feat$1gram$%s$%s$16to16$%s"%(self.corptable, self.correl_field, str(pocc).replace('.', '_'))

    def getWordGetter(self, lexicon_count_table=None):
        """Returns a FeatureGetter used for getting word counts. Usually used for group_freq_thresh.
 
        Parameters
        ----------
        lexicon_count_table : :obj:`str`, optional
            name of message table.
     
        Returns
        -------
        FeatureGetter
        """
        from .featureGetter import FeatureGetter
        if lexicon_count_table: dlac.warn(lexicon_count_table)
        wordTable = self.getWordTable() if not lexicon_count_table else lexicon_count_table

        assert mm.tableExists(self.corpdb, self.dbCursor, wordTable), "Need to create word table to use current functionality: %s" % wordTable
        return FeatureGetter(self.corpdb, self.corptable, self.correl_field, self.mysql_host,
                             self.message_field, self.messageid_field, self.encoding, self.use_unicode, 
                             self.lexicondb, featureTable=wordTable, wordTable = wordTable)
    
    def getWordGetterPOcc(self, pocc):
        """Returns a FeatureGetter for given p_occ filter values used for getting word counts. Usually used for group_freq_thresh.
 
        Parameters
        ----------
        pocc : float
            Value of p occurrence filter.
     
        Returns
        -------
        FeatureGetter
        """
        from .featureGetter import FeatureGetter
        wordTable = self.getWordTablePOcc(pocc)
        assert mm.tableExists(self.corpdb, self.dbCursor, wordTable, charset=self.encoding, use_unicode=self.use_unicode), "Need to create word table to use current functionality"
        return FeatureGetter(self.corpdb, self.corptable, self.correl_field, self.mysql_host,
                             self.message_field, self.messageid_field, self.encoding, self.use_unicode, 
                             self.lexicondb, featureTable=wordTable, wordTable = wordTable)

    def getGroupWordCounts(self, where = '', lexicon_count_table=None):
        """Get word counts for groups
 
        Parameters
        ----------
        where : :obj:`str`, optional
            Filter groups with sql-style call.
        lexicon_count_table : :obj:`str`, optional
            ?????
     
        Returns
        -------
        dict
            {group_id: sum(values)}
        """
        wordGetter = self.getWordGetter(lexicon_count_table)
        return dict(wordGetter.getSumValuesByGroup(where))

    def getTables(self, feat_table = False, like = None):
        """Returns a list of available tables.
 
        Parameters
        ----------
        feat_table : :obj:`str`, optional
            Indicator for listing feature tables

        like : :obj:`boolean`, optional
            Filter tables with sql-style call.
     
        Returns
        -------
        list
            A list of tables names
        """
        if feat_table:
            sql = """SHOW TABLES FROM %s LIKE 'feat$%%$%s$%s$%%' """ % (self.corpdb, self.corptable, self.correl_field)
        else:
            sql = """SHOW TABLES FROM %s where Tables_in_%s NOT LIKE 'feat%%' """ % (self.corpdb, self.corpdb)
            if isinstance(like, str): sql += """ AND Tables_in_%s like '%s'""" % (self.corpdb, like)
        return mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

    def describeTable(self, table_name):
        """
 
        Parameters
        ----------
        table_name : :obj:`str`
            Name of table to describe
     
        Returns
        -------
        Description of table (list of lists)
            
        """
        sql = """DESCRIBE %s""" % (table_name)
        return mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode)

    def viewTable(self, table_name):
        """
 
        Parameters
        ----------
        table_name : :obj:`str`
            Name of table to describe
     
        Returns
        -------
        First 5 rows of table (list of lists)
            
        """
        col_sql = """select column_name from information_schema.columns 
            where table_schema = '%s' and table_name='%s'""" % (self.corpdb, table_name)
        col_names = [col[0] for col in mm.executeGetList(self.corpdb, self.dbCursor, col_sql, charset=self.encoding, use_unicode=self.use_unicode)]
        sql = """SELECT * FROM %s LIMIT 10""" % (table_name)
        return [col_names] + list(mm.executeGetList(self.corpdb, self.dbCursor, sql, charset=self.encoding, use_unicode=self.use_unicode))


    def createRandomSample(self, percentage, random_seed = dlac.DEFAULT_RANDOM_SEED, where = ''):
        """Creates a new table from a random subetset of rows.
 
        Parameters
        ----------
        percentage : :obj:`float`, optional
            Percentage of random rows to keep

        random_seed : :obj:`int`, optional
            Filter groups with sql-style call.
     
        Returns
        -------
        list
            A list of feature tables names
        """
        new_table = self.corptable + "_rand"

        rows_sql = """SELECT count(*) from %s""" % (self.corptable)
        n_old_rows = mm.executeGetList1(self.corpdb, self.dbCursor, rows_sql, charset=self.encoding, use_unicode=self.use_unicode)[0]
        n_new_rows = round(percentage*n_old_rows)

        drop_sql = """DROP TABLE IF EXISTS %s""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, drop_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)

        create_sql = """CREATE TABLE %s LIKE %s""" % (new_table, self.corptable)
        mm.execute(self.corpdb, self.dbCursor, create_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        disable_sql = """ALTER TABLE %s DISABLE KEYS""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, disable_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        insert_sql = """INSERT INTO %s SELECT * FROM %s where RAND(%s) < %s""" % (new_table, self.corptable, random_seed, percentage*1.1)
        if where: insert_sql += " AND %s" % (where)
        insert_sql += " LIMIT %s" % (n_new_rows)
        mm.execute(self.corpdb, self.dbCursor, insert_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        enable_sql = """ALTER TABLE %s ENABLE KEYS""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, enable_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)

        return new_table

    def createCopiedTable(self, old_table, new_table, where = ''):
        """Creates a new table as a copy of an old table.
 
        Parameters
        ----------
        old_table: :obj:`string`, 
            name of the table to be copied

        new_table: :obj:`string`, 
            name of the new table

        Returns
        -------
        string
            new table name
        """
        #drop_sql = """DROP TABLE IF EXISTS %s""" % (new_table)
        #mm.execute(self.corpdb, self.dbCursor, drop_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)

        create_sql = """CREATE TABLE %s LIKE %s""" % (new_table, old_table)
        mm.execute(self.corpdb, self.dbCursor, create_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        disable_sql = """ALTER TABLE %s DISABLE KEYS""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, disable_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        insert_sql = """INSERT INTO %s SELECT * FROM %s""" % (new_table, old_table)
        if where: insert_sql += " WHERE %s" % (where)
        mm.execute(self.corpdb, self.dbCursor, insert_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)
        
        enable_sql = """ALTER TABLE %s ENABLE KEYS""" % (new_table)
        mm.execute(self.corpdb, self.dbCursor, enable_sql, warnQuery=True, charset=self.encoding, use_unicode=self.use_unicode)

        return new_table

    @staticmethod
    def makeBlackWhiteList(args_featlist, args_lextable, args_categories, args_lexdb, args_use_unicode):
        """?????
 
        Parameters
        ----------
        args_featlist : str
            ?????
        args_lextable : str
            ?????
        args_categories : str
            ?????
        args_lexdb : str
            ?????
        args_use_unicode : boolean
            ?????
     
        Returns
        -------
        newlist : list
            ?????
        """
        newlist = set()
        if args_use_unicode:
            print("making black or white list: [%s] [%s] [%s]" %([feat if isinstance(feat, str) else feat for feat in args_featlist], args_lextable, args_categories))
        else:
            print("making black or white list: [%s] [%s] [%s]" %([feat if isinstance(feat, str) else feat for feat in args_featlist], args_lextable, args_categories))
        if args_lextable:
            (conn, cur, dcur) = mm.dbConnect(args_lexdb, charset=dlac.DEF_ENCODING, use_unicode=args_use_unicode)
            sql = 'SELECT term FROM %s' % (args_lextable)
            if (len(args_categories) > 0) and args_categories[0] != '*':
                sql = 'SELECT term FROM %s WHERE category in (%s)'%(args_lextable, ','.join(['\''+str(x)+'\'' for x in args_categories]))

            rows = mm.executeGetList(args_lexdb, cur, sql, charset=dlac.DEF_ENCODING, use_unicode=args_use_unicode)
            for row in rows:
                newlist.add(row[0])
        elif args_featlist:
            for feat in args_featlist:
                if args_use_unicode:
                    feat = feat if isinstance(feat, str) else feat
                else:
                    feat = feat if isinstance(feat, str) else feat
                # newlist.add(feat.lower())
                if args_use_unicode:
                    newlist.add(feat.upper() if sum(map(str.isupper, feat)) > (len(feat)/2) else feat.lower())
                else:
                    newlist.add(feat.upper() if sum(map(str.isupper, feat)) > (len(feat)/2) else feat.lower())
        else:
            raise Exception('blacklist / whitelist flag specified without providing features.')
        newlist = [w.strip() for w in newlist]
        return newlist

#################################################################
### Main / Command-Line Processing:
##
#
if __name__ == "__main__":
    ##Argument Parser:
    print("dlaWorker.py Command-line Interface is Deprecated.\nUse dlatkInterface.py")
    sys.exit(0)




import sys
import os
import subprocess

_version = sys.version_info[0]

try:
  from setuptools import setup
except ImportError:
  if _version >= 3:
    sys.exit("Need setuptools to install dlatk for Python 3.x")
  from distutils.core import setup


DESCRIPTION = """DLATK is an end to end human text analysis package, specifically suited for social media and social scientific applications. It is written in Python 3 and developed by the World Well-Being Project at the University of Pennsylvania and Stony Brook University. """
LONG_DESCRIPTION = """
# Differential Language Analysis ToolKit

DLATK is an end to end human text analysis package, specifically suited for social media and social scientific applications. It is written in Python 3 and developed by the World Well-Being Project at the University of Pennsylvania and Stony Brook University. 

It contains:

- feature extraction
- part-of-speech tagging
- correlation
- prediction and classification
- mediation 
- dimensionality reduction and clustering
- wordcloud visualization

DLATK can utilize:

- [Mallet](http://mallet.cs.umass.edu/) for creating LDA topics
- [Stanford Parser](http://nlp.stanford.edu/software/lex-parser.shtml) 
- [CMU's TweetNLP](http://www.cs.cmu.edu/~ark/TweetNLP/) 
- [pandas](http://pandas.pydata.org/) dataframe output

## Installation

DLATK is available via any of four popular installation platforms: conda, pip, github, or Docker:

#### New to installing Python packages?
It is recommended that you see the [full installation instructions](http://dlatk.wwbp.org/install.html#dependencies). 

### 1. conda
```sh
conda install -c wwbp dlatk
```

### 2. pip
```sh
pip install dlatk
```

### 3. GitHub
```sh
git clone https://github.com/dlatk/dlatk.git
cd dlatk
python setup.py install
```

### 4. Docker
Detailed Docker install instructions [here](http://dlatk.wwbp.org/tutorials/tut_docker.html).

```sh
docker run --name mysql_v5  --env MYSQL_ROOT_PASSWORD=my-secret-pw --detach mysql:5.5
docker run -it --rm --name dlatk_docker --link mysql_v5:mysql dlatk/dlatk bash
```

- [DLATK at DockerHub](https://hub.docker.com/r/dlatk/dlatk/)
- [DockerFile on GitHub](https://github.com/dlatk/dlatk-docker)

## Dependencies
- [mysqlclient](https://github.com/PyMySQL/mysqlclient-python)
- [NumPy](http://www.numpy.org)
- [scikit-learn](http://www.scikit-learn.org/)
- [SciPy](http://www.scipy.org/)
- [statsmodels](http://www.statsmodels.org/)

See the [full installation instructions](http://dlatk.wwbp.org/install.html#dependencies)
for recommended and optional dependencies.

## Documentation

The documentation for the latest release is at [dlatk.wwbp.org](dlatk.wwbp.org).

## Citation

If you use DLATK in your work please cite the following [paper](http://aclweb.org/anthology/D17-2010):

```
@InProceedings{DLATKemnlp2017,
  author =  "Schwartz, H. Andrew
    and Giorgi, Salvatore
    and Sap, Maarten
    and Crutchley, Patrick
    and Eichstaedt, Johannes
    and Ungar, Lyle",
  title =   "DLATK: Differential Language Analysis ToolKit",
  booktitle =   "Proceedings of the 2017 Conference on Empirical Methods in Natural Language Processing: System Demonstrations",
  year =  "2017",
  publisher =   "Association for Computational Linguistics",
  pages =   "55--60",
  location =  "Copenhagen, Denmark",
  url =   "http://aclweb.org/anthology/D17-2010"
}

```

## License

Licensed under a [GNU General Public License v3 (GPLv3)](https://www.gnu.org/licenses/gpl-3.0.en.html)

## Background

Developed by the [World Well-Being Project](http://www.wwbp.org) based out of the University of Pennsylvania and Stony Brook University.
"""
DISTNAME = 'dlatk'
PACKAGES = ['dlatk',
  'dlatk.lib',
  'dlatk.lexicainterface',
  'dlatk.mysqlmethods',
  'dlatk.tools',
]
LICENSE = 'GNU General Public License v3 (GPLv3)'
AUTHOR = "H. Andrew Schwartz, Salvatore Giorgi, Maarten Sap, Patrick Crutchley, Lukasz Dziurzynski and Megha Agrawal"
EMAIL = "has@cs.stonybrook.edu, sgiorgi@sas.upenn.edu"
MAINTAINER = "Salvatore Giorgi, H. Andrew Schwartz"
MAINTAINER_EMAIL = "sgiorgi@sas.upenn.edu, has@cs.stonybrook.edu"
URL = "http://dlatk.wwbp.org"
DOWNLOAD_URL = 'https://github.com/dlatk/dlatk'
CLASSIFIERS = [
  'Environment :: Console',
  'Natural Language :: English',
  'Intended Audience :: End Users/Desktop',
  'Intended Audience :: Developers',
  'Intended Audience :: Science/Research',
  'Programming Language :: Python',
  'Programming Language :: Python :: 3',
  'Programming Language :: Python :: 3.5',
  'Topic :: Scientific/Engineering',
]
VERSION = '1.1.8'
PACKAGE_DATA = {
  'dlatk': ['data/*.sql'],
  'dlatk.lib': ['lib/meloche_bd.ttf', 'lib/oval_big_mask.png', 'lib/oval_mask.png'],
}
INCLUDE_PACKAGE_DATA = True
SETUP_REQUIRES = [
  'numpy', 
]
INSTALL_REQUIRES = [
  'matplotlib>=1.3.1', 
  'mysqlclient', 
  'nltk>=3.2.1', 
  'numpy', 
  'pandas>=0.20.3', 
  'patsy>=0.2.1', 
  'python-dateutil>=2.5.0', 
  'scikit-learn==0.18.2', 
  'scipy>=0.19.1', 
  'SQLAlchemy>=1.0.13', 
  'statsmodels>=0.8.0', 
]
EXTRAS_REQUIRE = {
  'image': ['image'],
  'jsonrpclib-pelix': ['jsonrpclib-pelix>=0.2.8'],
  'langid': ['langid>=1.1.4'],
  'rpy2': ['rpy2'],
  'simplejson': ['simplejson>=3.3.1'],
  'textstat': ['textstat>=0.6.1'],
  'wordcloud':  ['wordcloud==1.1.3'],
}

SCRIPTS = ['dlatkInterface.py']

if __name__ == "__main__":

  setup(name=DISTNAME,
      author=AUTHOR,
      author_email=EMAIL, 
      version=VERSION,
      packages=PACKAGES,
      package_data=PACKAGE_DATA,
      include_package_data=INCLUDE_PACKAGE_DATA,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      license=LICENSE,
      url=URL,
      download_url=DOWNLOAD_URL,
      classifiers=CLASSIFIERS,
      setup_requires=SETUP_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      install_requires=INSTALL_REQUIRES,
      scripts = SCRIPTS,
      long_description_content_type ='text/markdown',
  )




#!/usr/bin/env python
#########################################
from numpy import log2, isnan
import csv
import os, sys
sys.path.append(os.path.dirname(os.path.realpath(__file__)).replace("/dlatk/LexicaInterface",""))

from dlatk.featureExtractor import FeatureExtractor
from dlatk import dlaConstants as dlac

from json import loads

class TopicExtractor(FeatureExtractor):

    def __init__(self, corpdb=dlac.DEF_CORPDB, corptable=dlac.DEF_CORPTABLE, correl_field=dlac.DEF_CORREL_FIELD, 
                 mysql_host = "localhost", message_field=dlac.DEF_MESSAGE_FIELD, messageid_field=dlac.DEF_MESSAGEID_FIELD, 
                 encoding=dlac.DEF_ENCODING, use_unicode=dlac.DEF_UNICODE_SWITCH, ldaMsgTable =dlac.DEF_LDA_MSG_TABLE):
        super(TopicExtractor, self).__init__(corpdb, corptable, correl_field, mysql_host, message_field, messageid_field, encoding, use_unicode)
        self.ldaMsgTable = ldaMsgTable


    def createDistributions(self, filename=None):

        if not filename:
            filename = self.ldaMsgTable.replace('$', '.')

        word_freq = dict() #stores frequencies of each word type
        topic_freq = dict() #stores frequencies of each topic
        topic_word_freq = dict() #stores frequencies words being in a particular topics
        all_freq = 0 # stores the number of word encountered

        print("[Generating whitelist]")
        #whitelist = set(self.getWordWhiteList())
        whitelist = None
        print("[Done]")

        #update freqs based on 1 msg at a time
        print("[Finding lda messages]")
        msgs = 0
        ssMsgCursor = self.getMessages(messageTable = self.ldaMsgTable)#, where = "user_id < 2000000")
        for messageRow in ssMsgCursor: #important that this just read one line at a time
            message_id = messageRow[0]
            topicsEncoded = messageRow[1]
            if topicsEncoded:
                msgs+=1
                if msgs % 5000 == 0: #progress update
                    print(("Messages Read: %dk" % int(msgs/1000)))

                #print "encoded: %s " % str(topicsEncoded)
                wordTopics = loads(topicsEncoded)
                #print "decoded: %s" % str(wordTopics)

                for wt in wordTopics:
                   (word, topic) = (wt['term'], wt['topic_id'])
                   #update word frequencies:
                   all_freq += 1
                   if not topic in topic_freq: #this could go inside the whitelist check also, with minimal change to output
                       topic_freq[topic] = 1
                       topic_word_freq[topic] = dict()
                   else:
                       topic_freq[topic] += 1

                   #if word in whitelist:
                   if not word in word_freq:
                       word_freq[word] = 1
                   else:
                       word_freq[word] += 1
                   if not word in topic_word_freq[topic]:
                        topic_word_freq[topic][word] = 1
                   else:
                        topic_word_freq[topic][word] += 1

        #COMPUTE DISTRIBUTIONS:
        print("[Computing Distributions]")
        pTopicGivenWords = dict()
        likelihoods = dict()
        log_likelihoods = dict()
        pWordGivenTopics = dict()

        for topic in topic_freq:
            pTopic = topic_freq[topic] / float(all_freq)
            pTopicGivenWords[topic] = dict()
            likelihoods[topic] = dict()
            log_likelihoods[topic] = dict()
            pWordGivenTopics[topic] = dict()

            for word in topic_word_freq[topic]:
                pWord = word_freq[word] / float(all_freq)
                pWordTopic = topic_word_freq[topic][word] / float(all_freq)
                pTopicGivenWords[topic][word] = pWordTopic / pWord
                likelihoods[topic][word] = pWordTopic / (pWord * pTopic)
                log_likelihoods[topic][word] = log2(1+likelihoods[topic][word])
                pWordGivenTopics[topic][word] = pWordTopic / pTopic

        #print pTopicGivenWords to file
        self.printDistToCSV(pTopicGivenWords, filename+'.topicGivenWord.csv')
        #print likelihoods to file
        #self.printDistToCSV(likelihoods, filename+'.lik.csv')
        #print log_likelihoods to file
        self.printDistToCSV(log_likelihoods, filename+'.loglik.csv')
        #print word given topic
        self.printDistToCSV(pWordGivenTopics, filename+'.wordGivenTopic.csv')

        #threshold:
        newLLs = dict()
        for topic, wordDist in log_likelihoods.items():
            newLLs[topic] = dict()
            sortedWordValues = sorted(list(wordDist.items()), key = lambda w_v: w_v[1] if not isnan(w_v[1]) else -1000, reverse = True)
            threshold = 0
            if len(sortedWordValues) > 1:
                threshold = 0.50 * sortedWordValues[0][1]
            for word, value in sortedWordValues:
                if value > threshold:
                    newLLs[topic][word] = topic_word_freq[topic][word]
        self.printDistToCSV(newLLs, filename+'.freq.threshed50.loglik.csv')

        #TODO: print topics to tables:
        #id, topic, term, pcond, lik, loglik


    @staticmethod
    def printDistToCSV(dist, fileName):
        print("[Writing Distribution CSV to %s]" %fileName)
        csvWriter = csv.writer(open(fileName, 'w'))
        csvWriter.writerow(['topic_id', 'word1', 'word1_score', 'word2', 'word2_score', '...'])
        for topic in sorted(list(dist.keys()), key = lambda k: int(k) if str(k).isdigit() else k):
            wordDist = dist[topic]
            row = [topic]
            for wordValue in sorted(list(wordDist.items()), key = lambda w_v1: w_v1[1] if not isnan(w_v1[1]) else -1000, reverse = True):
                row.extend(wordValue)
            csvWriter.writerow(row)


    def getWordWhiteList(self, pocc = 0.01):
        wg = self.getWordGetterPOcc(pocc)
        return wg.getDistinctFeatures()


    def addLDAFeatTable(self, ldaMessageTable, tableName = None, valueFunc = lambda d: d):
        """Creates feature tuples (correl_field, feature, values) table where features are ngrams"""
        """Optional ValueFunc program scales that features by the function given"""
        #CREATE TABLE:
        featureName =  'lda'+'_'+ldaMessageTable.split('$')[1]
        featureTableName = self.createFeatureTable(featureName, 'SMALLINT UNSIGNED', 'INTEGER', tableName, valueFunc)

        #SELECT / LOOP ON CORREL FIELD FIRST:
        usql = """SELECT %s FROM %s GROUP BY %s""" % (
            self.correl_field, self.corptable, self.correl_field)
        msgs = 0#keeps track of the number of messages read
        cfRows = self._executeGetList(usql)#SSCursor woudl be better, but it loses connection
        dlac.warn("finding messages for %d '%s's"%(len(cfRows), self.correl_field))
        self._disableTableKeys(featureTableName)#for faster, when enough space for repair by sorting
        for cfRow in cfRows: 
            cf_id = cfRow[0]
            mids = set() #currently seen message ids
            freqs = dict() #holds frequency of n-grams
            totalInsts = 0 #total number of (non-distinct) topics

            #grab topics by messages for that cf:
            for messageRow in self.getMessagesForCorrelField(cf_id, messageTable = ldaMessageTable):
                message_id = messageRow[0]
                topicsEncoded = messageRow[1]
                if not message_id in mids and topicsEncoded:
                    msgs+=1
                    if msgs % PROGRESS_AFTER_ROWS == 0: #progress update
                        dlac.warn("Messages Read: %dk" % int(msgs/1000))
                    #print topicsEncoded
                    topics = json.loads(topicsEncoded)

                    for topic in topics:
                        totalInsts += 1
                        topicId = topic['topic_id']
                        if not topicId in freqs:
                            freqs[topicId] = 1
                        else: 
                            freqs[topicId] += 1
                    mids.add(message_id)

            #write topic to database (no need for "REPLACE" because we are creating the table)
            wsql = """INSERT INTO """+featureTableName+""" (group_id, feat, value, group_norm) values ('"""+str(cf_id)+"""', %s, %s, %s)"""
            totalInsts = float(totalInsts) #to avoid casting each time below
            rows = [(k, v, valueFunc((v / totalInsts))) for k, v in freqs.items() ] #adds group_norm and applies freq filter
            self._executeWriteMany(wsql, rows)
        
        dlac.warn("Done Reading / Inserting.")

        dlac.warn("Adding Keys (if goes to keycache, then decrease MAX_TO_DISABLE_KEYS or run myisamchk -n).")
        self._enableTableKeys(featureTableName)#rebuilds keys
        dlac.warn("Done\n")
        return featureTableName;


